#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Automatically fix lintian issues."""

from debian.changelog import Changelog
import errno
import io
import os
import re
import shutil
import subprocess
import sys
import tempfile
import time
import traceback
import warnings

from breezy import ui

import breezy.bzr  # noqa: F401
import breezy.git  # noqa: F401
from breezy.clean_tree import (
    iter_deletables,
    )
from breezy.commit import NullCommitReporter
from breezy.errors import NoSuchFile
from breezy.osutils import is_inside
from breezy.rename_map import RenameMap
from breezy.trace import note, warning
from breezy.transform import revert

from debian.deb822 import Deb822


from .reformatting import FormattingUnpreservable


__version__ = (0, 60)
version_string = '.'.join(map(str, __version__))
SUPPORTED_CERTAINTIES = ['certain', 'confident', 'likely', 'possible', None]
DEFAULT_MINIMUM_CERTAINTY = 'certain'
USER_AGENT = 'lintian-brush/' + version_string
# Too aggressive?
DEFAULT_URLLIB_TIMEOUT = 3


class NoChanges(Exception):
    """Script didn't make any changes."""


class NotCertainEnough(NoChanges):
    """Script made changes but with too low certainty."""

    def __init__(self, certainty, minimum_certainty):
        self.certainty = certainty
        self.minimum_certainty = minimum_certainty


class FixerFailed(Exception):
    """Base class for fixer script failures."""

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.args == other.args


class UnsupportedCertainty(Exception):
    """Unsupported certainty."""


class FixerScriptFailed(FixerFailed):
    """Script failed to run."""

    def __init__(self, path, returncode, errors):
        self.path = path
        self.returncode = returncode
        self.errors = errors

    def __str__(self):
        return ("Script %s failed with exit code: %d\n%s\n" % (
                self.path, self.returncode,
                self.errors))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return (
            self.path == other.path and
            self.returncode == other.returncode and
            self.errors == other.errors)


class DescriptionMissing(Exception):
    """The fixer script did not provide a description on stdout."""


class NotDebianPackage(Exception):
    """The specified directory does not contain a Debian package."""

    def __init__(self, tree, path):
        super(NotDebianPackage, self).__init__(tree.abspath(path))


class PendingChanges(Exception):
    """The directory has pending changes."""

    def __init__(self, tree):
        super(PendingChanges, self).__init__(tree.basedir)


class FixerResult(object):
    """Result of a fixer run."""

    def __init__(self, description, fixed_lintian_tags=[],
                 certainty=None, patch_name=None,
                 revision_id=None):
        self.description = description
        self.fixed_lintian_tags = fixed_lintian_tags
        self.certainty = certainty
        self.patch_name = patch_name
        self.revision_id = revision_id

    def __repr__(self):
        return ("%s(%r, fixed_lintian_tags=%r, certainty=%r, patch_name=%r, "
                "revision_id=%r)") % (
                self.__class__.__name__,
                self.description, self.fixed_lintian_tags, self.certainty,
                self.patch_name, self.revision_id)

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return ((self.description == other.description) and
                (self.fixed_lintian_tags == other.fixed_lintian_tags) and
                (self.certainty == other.certainty) and
                (self.patch_name == other.patch_name) and
                (self.revision_id == other.revision_id))


class Fixer(object):
    """A Fixer script.

    The `lintian_tags` attribute contains the name of the lintian tags this
    fixer addresses.
    """

    def __init__(self, name, lintian_tags=None):
        self.name = name
        self.lintian_tags = lintian_tags or []

    def run(self, basedir, package, current_version, compat_release,
            minimum_certainty=None, trust_package=False,
            allow_reformatting=False, net_access=True, opinionated=False,
            diligence=0):
        """Apply this fixer script.

        Args:
          basedir: Directory in which to run
          package: Name of the source package
          current_version: The version of the package that is being created or
            updated
          compat_release: Compatibility level (a Debian release name)
          trust_package: Whether to run code from the package
          allow_reformatting: Allow reformatting of files that are being
            changed
          opinionated: Whether to be opinionated
          diligence: Level of diligence
        Returns:
          A FixerResult object
        """
        raise NotImplementedError(self.run)


def parse_script_fixer_output(text):
    """Parse the output from a script fixer."""
    lines = []
    fixed_tags = []
    certainty = None
    patch_name = None
    for line in text.splitlines():
        # TODO(jelmer): Do this in a slighly less hackish manner
        try:
            (key, value) = line.split(':', 1)
        except ValueError:
            lines.append(line)
        else:
            if key == 'Fixed-Lintian-Tags':
                fixed_tags = value.strip().split(',')
            elif key == 'Certainty':
                certainty = value.strip()
            elif key == 'Patch-Name':
                patch_name = value.strip()
            else:
                lines.append(line)
    if certainty not in SUPPORTED_CERTAINTIES:
        raise UnsupportedCertainty(certainty)
    return FixerResult('\n'.join(lines), fixed_tags, certainty, patch_name)


def determine_env(package, current_version, compat_release, minimum_certainty,
                  trust_package, allow_reformatting, net_access, opinionated,
                  diligence):
    env = dict(os.environ.items())
    env['PACKAGE'] = package
    env['CURRENT_VERSION'] = str(current_version)
    env['COMPAT_RELEASE'] = compat_release
    env['MINIMUM_CERTAINTY'] = minimum_certainty
    env['TRUST_PACKAGE'] = 'true' if trust_package else 'false'
    env['REFORMATTING'] = ('allow' if allow_reformatting else 'disallow')
    env['NET_ACCESS'] = ('allow' if net_access else 'disallow')
    env['OPINIONATED'] = ('yes' if opinionated else 'no')
    env['DILIGENCE'] = str(diligence)
    return env


class PythonScriptFixer(Fixer):
    """A fixer that is implemented as a python script.

    This gets used just for Python scripts, and significantly speeds
    things up because it prevents starting a new Python interpreter
    for every fixer.
    """

    def __init__(self, name, lintian_tags, script_path):
        super(PythonScriptFixer, self).__init__(name, lintian_tags)
        self.script_path = script_path

    def __repr__(self):
        return "<%s(%r)>" % (self.__class__.__name__, self.name)

    def run(self, basedir, package, current_version, compat_release,
            minimum_certainty=DEFAULT_MINIMUM_CERTAINTY,
            trust_package=False, allow_reformatting=False,
            net_access=True, opinionated=False, diligence=0):
        env = determine_env(
            package=package,
            current_version=current_version,
            compat_release=compat_release,
            minimum_certainty=minimum_certainty,
            trust_package=trust_package,
            allow_reformatting=allow_reformatting,
            net_access=net_access,
            opinionated=opinionated,
            diligence=diligence)
        try:
            old_env = os.environ
            old_stderr = sys.stderr
            old_stdout = sys.stdout
            sys.stderr = io.StringIO()
            sys.stdout = io.StringIO()
            os.environ = env
            old_cwd = os.getcwd()
            try:
                os.chdir(basedir)
                global_vars = {
                    "__file__": self.script_path,
                    "__name__": "__main__",
                    }
                with open(self.script_path, 'r') as f:
                    code = compile(f.read(), self.script_path, 'exec')
                    exec(code, global_vars)
            except FormattingUnpreservable:
                raise
            except SystemExit as e:
                retcode = e.code
            except BaseException as e:
                traceback.print_exception(
                    type(e), e, e.__traceback__, file=sys.stderr)
                raise FixerScriptFailed(
                    self.script_path, 1,
                    sys.stderr.getvalue())
            else:
                retcode = 0
            description = sys.stdout.getvalue()
            err = sys.stderr.getvalue()
        finally:
            os.environ = old_env
            sys.stderr = old_stderr
            sys.stdout = old_stdout
            os.chdir(old_cwd)

        if retcode == 2:
            raise NoChanges()
        if retcode != 0:
            raise FixerScriptFailed(self.script_path, retcode, err)

        return parse_script_fixer_output(description)


class ScriptFixer(Fixer):
    """A fixer that is implemented as a shell/python/etc script."""

    def __init__(self, name, lintian_tags, script_path):
        super(ScriptFixer, self).__init__(name, lintian_tags)
        self.script_path = script_path

    def __repr__(self):
        return "<ScriptFixer(%r)>" % self.name

    def run(self, basedir, package, current_version, compat_release,
            minimum_certainty=DEFAULT_MINIMUM_CERTAINTY,
            trust_package=False, allow_reformatting=False,
            net_access=True, opinionated=False, diligence=0):
        env = determine_env(
            package=package,
            current_version=current_version,
            compat_release=compat_release,
            minimum_certainty=minimum_certainty,
            trust_package=trust_package,
            allow_reformatting=allow_reformatting,
            net_access=net_access,
            opinionated=opinionated,
            diligence=diligence)
        with tempfile.SpooledTemporaryFile() as stderr:
            p = subprocess.Popen(self.script_path, cwd=basedir,
                                 stdout=subprocess.PIPE, stderr=stderr,
                                 env=env)
            (description, err) = p.communicate("")
            if p.returncode == 2:
                raise NoChanges()
            if p.returncode != 0:
                stderr.seek(0)
                raise FixerScriptFailed(
                        self.script_path, p.returncode,
                        stderr.read().decode('utf-8', 'replace'))
        return parse_script_fixer_output(description.decode('utf-8'))


def find_fixers_dir():
    """Find the local directory with lintian fixer scripts."""
    local_dir = os.path.abspath(os.path.join(
        os.path.dirname(__file__), '..', 'fixers'))
    if os.path.isdir(local_dir):
        return local_dir
    import pkg_resources
    resource_dir = pkg_resources.resource_filename(
        __name__, 'lintian-brush/fixers')
    if os.path.isdir(resource_dir):
        return resource_dir
    # Urgh.
    return '/usr/share/lintian-brush/fixers'


def read_desc_file(path):
    """Read a description file.

    Args:
      path: Path to read from.
    Yields:
      Fixer objects
    """
    dirname = os.path.dirname(path)
    with open(path, 'r') as f:
        for paragraph in Deb822.iter_paragraphs(f):
            name = os.path.splitext(paragraph['Fix-Script'])[0]
            script_path = os.path.join(dirname, paragraph['Fix-Script'])
            if 'Lintian-Tags' in paragraph:
                tags = [tag.strip()
                        for tag in paragraph['Lintian-Tags'].split(',')]
            else:
                tags = []
            if script_path.endswith('.py'):
                yield PythonScriptFixer(name, tags, script_path)
            else:
                yield ScriptFixer(name, tags, script_path)


def select_fixers(fixers, names, exclude=None):
    """Select fixers by name, from a list.

    Args:
      fixers: List of Fixer objects
      names: List of names to select
      exclude: List of names to exclude
    Raises:
      KeyError: if one of the names did not exist
    """
    names = set(names)
    if exclude:
        for name in exclude:
            if name not in names:
                raise KeyError(name)
            names.remove(name)
    available = set([f.name for f in fixers])
    missing = names - available
    if missing:
        raise KeyError(missing.pop())
    # Preserve order
    return [f for f in fixers if f.name in names]


def available_lintian_fixers(fixers_dir=None):
    """Return a list of available lintian fixers.

    Args:
      fixers_dir: Fixers directory to browse
    Returns:
      Iterator over Fixer objects
    """
    if fixers_dir is None:
        fixers_dir = find_fixers_dir()
    for n in os.listdir(fixers_dir):
        if not n.endswith(".desc"):
            continue
        for fixer in read_desc_file(os.path.join(fixers_dir, n)):
            yield fixer


def increment_version(v):
    """Increment a version number.

    For native packages, increment the main version number.
    For other packages, increment the debian revision.

    Args:
        v: Version to increment (modified in place)
    """
    if v.debian_revision is not None:
        v.debian_revision = re.sub(
                '\\d+$', lambda x: str(int(x.group())+1), v.debian_revision)
    else:
        v.upstream_version = re.sub(
                '\\d+$', lambda x: str(int(x.group())+1), v.upstream_version)


def delete_items(deletables, dry_run=False):
    """Delete files in the deletables iterable"""
    def onerror(function, path, excinfo):
        """Show warning for errors seen by rmtree.
        """
        # Handle only permission error while removing files.
        # Other errors are re-raised.
        if function is not os.remove or excinfo[1].errno != errno.EACCES:
            raise
        warnings.warn('unable to remove %s' % path)
    for path, subp in deletables:
        if os.path.isdir(path):
            shutil.rmtree(path, onerror=onerror)
        else:
            try:
                os.unlink(path)
            except OSError as e:
                # We handle only permission error here
                if e.errno != errno.EACCES:
                    raise e
                warnings.warn(
                    'unable to remove "{0}": {1}.'.format(path, e.strerror))


def get_committer(tree):
    """Get the committer string for a tree.

    Args:
      tree: A Tree object
    Returns:
      A committer string
    """
    # TODO(jelmer): Perhaps this logic should be in Breezy?
    if getattr(tree.branch.repository, '_git', None):
        cs = tree.branch.repository._git.get_config_stack()
        user = os.environ.get("GIT_COMMITTER_NAME")
        email = os.environ.get("GIT_COMMITTER_EMAIL")
        if user is None:
            try:
                user = cs.get(("user", ), "name")
            except KeyError:
                user = None
            else:
                user = user.decode('utf-8')
        if email is None:
            try:
                email = cs.get(("user", ), "email")
            except KeyError:
                email = None
            else:
                email = email.decode('utf-8')
        if user and email:
            return user + " <" + email + ">"
        from breezy.config import GlobalStack
        return GlobalStack().get('email')
    else:
        config = tree.branch.get_config_stack()
        return config.get('email')


def only_changes_last_changelog_block(tree, changelog_path, changes):
    """Check whether the only change in a tree is to the last changelog entry.

    Args:
      tree: Tree to analyze
      changelog_path: Path to the changelog file
      changes: Changes in the tree
    Returns:
      boolean
    """
    basis_tree = tree.basis_tree()
    with tree.lock_read(), basis_tree.lock_read():
        for change in changes:
            try:
                change_paths = change.path
            except AttributeError:  # brz < 3.1
                change_paths = change[1]
            if change_paths == ('', ''):
                continue
            if change_paths != (changelog_path, changelog_path):
                return False
            break
        else:
            return False
        new_cl = Changelog(tree.get_file_text(changelog_path))
        old_cl = Changelog(basis_tree.get_file_text(changelog_path))
        if old_cl.distributions != "UNRELEASED":
            return False
        del new_cl._blocks[0]
        del old_cl._blocks[0]
        return str(new_cl) == str(old_cl)


def reset_tree(local_tree, dirty_tracker=None, subpath='.'):
    """Reset a tree back to its basis tree.

    This will leave ignored and detritus files alone.

    Args:
      local_tree: tree to work on
      dirty_tracker: Optional dirty tracker
    """
    if dirty_tracker and not dirty_tracker.is_dirty():
        return
    revert(local_tree, local_tree.branch.basis_tree(),
           [subpath] if subpath != '.' else None)
    deletables = list(iter_deletables(
        local_tree, unknown=True, ignored=False, detritus=False))
    delete_items(deletables)


def certainty_sufficient(actual_certainty, minimum_certainty):
    """Check if the actual certainty is sufficient.

    Args:
      actual_certainty: Actual certainty with which changes were made
      minimum_certainty: Minimum certainty to keep changes
    Returns:
      boolean
    """
    actual_confidence = certainty_to_confidence(actual_certainty)
    if actual_confidence is None:
        # Actual confidence is unknown.
        # TODO(jelmer): Should we really be ignoring this?
        return True
    minimum_confidence = certainty_to_confidence(minimum_certainty)
    if minimum_confidence is None:
        return True
    return actual_confidence <= minimum_confidence


def check_clean_tree(local_tree):
    """Check that a tree is clean and has no pending changes or unknown files.

    Args:
      local_tree: The tree to check
    Raises:
      PendingChanges: When there are pending changes
    """
    # Just check there are no changes to begin with
    if local_tree.has_changes():
        raise PendingChanges(local_tree)
    if list(local_tree.unknowns()):
        raise PendingChanges(local_tree)


def add_changelog_entry(tree, path, summary, qa=False):
    """Add a changelog entry.

    Args:
      tree: Tree to edit
      path: Path to the changelog file
      summary: Entry to add
    """
    args = ["dch", "--no-auto-nmu"]
    if qa:
        args.append('--qa')
    p = subprocess.Popen(
        args + [summary],
        cwd=tree.abspath(os.path.dirname(os.path.dirname(path))),
        stdin=subprocess.PIPE, stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)
    (stdout, stderr) = p.communicate(b"\n")
    if stdout:
        raise ValueError(stdout)
    if stderr:
        end = b'dch: Did you see that warning?  Press RETURN to continue...\n'
        if stderr.endswith(end):
            stderr = stderr[:-len(end)]
        for line in stderr.splitlines():
            warning('%s', line)


def gbp_conf_has_dch_section(tree, path=''):
    try:
        gbp_conf_path = os.path.join(path, 'debian/gbp.conf')
        gbp_conf_text = tree.get_file_text(gbp_conf_path)
    except NoSuchFile:
        return False
    try:
        import configparser
    except ImportError:
        return None
    else:
        parser = configparser.ConfigParser(defaults={}, strict=False)
        parser.read_string(
            gbp_conf_text.decode('utf-8', errors='replace'), gbp_conf_path)
        return parser.has_section('dch')


def all_sha_prefixed(cl):
    sha_prefixed = 0
    for change in cl.changes():
        if not change.startswith('  * '):
            continue
        if re.match(r'  \* \[[0-9a-f]{7}\] ', change):
            sha_prefixed += 1
        else:
            return False
    return (sha_prefixed > 0)


_changelog_policy_noted = False


def guess_update_changelog(tree, path='', cl=None):
    """Guess whether the changelog should be updated.

    Args:
      tree: Tree to edit
      path: Path to packaging in tree
    """
    global _changelog_policy_noted
    if gbp_conf_has_dch_section(tree, path):
        if not _changelog_policy_noted:
            note('Assuming changelog does not need to be updated, '
                 'since there is a [dch] section in gbp.conf.')
            _changelog_policy_noted = True
        return False

    # TODO(jelmes): Do something more clever here, perhaps looking at history
    # of the changelog file?
    changelog_path = os.path.join(path, 'debian/changelog')
    if cl is None:
        try:
            with tree.get_file(changelog_path) as f:
                cl = Changelog(f, max_blocks=1)
        except NoSuchFile:
            cl = None

    if cl and all_sha_prefixed(cl[0]):
        return False

    return True


def has_non_debian_changes(changes):
    for change in changes:
        try:
            change_paths = change.path
        except AttributeError:  # brz < 3.1
            change_paths = change[1]
        for path in change_paths:
            if path and not is_inside('debian', path):
                return True
    return False


def run_lintian_fixer(local_tree, fixer, committer=None,
                      update_changelog=None, compat_release=None,
                      minimum_certainty=None, trust_package=False,
                      allow_reformatting=False, dirty_tracker=None,
                      subpath='.', net_access=True, opinionated=None,
                      diligence=0):
    """Run a lintian fixer on a tree.

    Args:
      local_tree: WorkingTree object
      fixer: Fixer object to apply
      committer: Optional committer (name and email)
      update_changelog: Whether to add a new entry to the changelog
      compat_release: Minimum release that the package should be usable on
        (e.g. 'stable' or 'unstable')
      minimum_certainty: How certain the fixer should be
        about its changes.
      trust_package: Whether to run code from the package if necessary
      allow_reformatting: Whether to allow reformatting of changed files
      dirty_tracker: Optional object that can be used to tell if the tree
        has been changed.
      subpath: Path in tree to operate on
      net_access: Whether to allow accessing external services
      opinionated: Whether to be opinionated
      diligence: Level of diligence
    Returns:
      tuple with set of FixerResult, summary of the changes
    """
    if subpath == '.':
        changelog_path = 'debian/changelog'
    else:
        changelog_path = os.path.join(subpath, 'debian/changelog')

    try:
        with local_tree.get_file(changelog_path) as f:
            cl = Changelog(f, max_blocks=1)
    except NoSuchFile:
        raise NotDebianPackage(local_tree, subpath)
    package = cl.package
    if cl.distributions == 'UNRELEASED':
        current_version = cl.version
    else:
        current_version = cl.version
        increment_version(current_version)
    if compat_release is None:
        compat_release = 'sid'
    if minimum_certainty is None:
        minimum_certainty = DEFAULT_MINIMUM_CERTAINTY
    try:
        result = fixer.run(
            local_tree.abspath(subpath),
            package=package,
            current_version=current_version,
            compat_release=compat_release,
            minimum_certainty=minimum_certainty,
            trust_package=trust_package,
            allow_reformatting=allow_reformatting,
            net_access=net_access,
            opinionated=opinionated,
            diligence=diligence)
    except BaseException:
        reset_tree(local_tree, dirty_tracker, subpath)
        raise
    if not certainty_sufficient(result.certainty, minimum_certainty):
        reset_tree(local_tree, dirty_tracker, subpath)
        raise NotCertainEnough(result.certainty, minimum_certainty)
    if dirty_tracker:
        relpaths = dirty_tracker.relpaths()
        # Sort paths so that directories get added before the files they
        # contain (on VCSes where it matters)
        local_tree.add(
            [p for p in sorted(relpaths)
             if local_tree.has_filename(p) and not
                local_tree.is_ignored(p)])
        specific_files = [
            p for p in relpaths
            if local_tree.is_versioned(p)]
        if not specific_files:
            raise NoChanges("Script didn't make any changes")
    else:
        local_tree.smart_add([local_tree.abspath(subpath)])
        specific_files = [subpath] if subpath != '.' else None

    basis_tree = local_tree.basis_tree()
    if local_tree.supports_setting_file_ids():
        RenameMap.guess_renames(
            basis_tree, local_tree, dry_run=False)

    changes = list(local_tree.iter_changes(
        basis_tree, specific_files=specific_files,
        want_unversioned=False, require_versioned=True))

    if len(local_tree.get_parent_ids()) <= 1 and not changes:
        raise NoChanges("Script didn't make any changes")

    if not result.description:
        raise DescriptionMissing()

    summary = result.description.splitlines()[0]

    # If there are upstream changes in a non-native package, perhaps
    # export them to debian/patches
    if has_non_debian_changes(changes) and current_version.debian_revision:
        # TODO(jelmer): Apply all patches before generating a diff.
        reset_tree(local_tree, dirty_tracker, subpath)
        raise NoChanges("Creating upstream patches not supported yet")

        from .patches import move_upstream_changes_to_patch
        try:
            specific_files, patch_name = move_upstream_changes_to_patch(
                local_tree, subpath, result.patch_name, result.description,
                dirty_tracker)
        except FileExistsError as e:
            raise NoChanges('patch path %s already exists\n' % e.args[0])

        summary = 'Add patch %s: %s' % (patch_name, summary)

    if update_changelog is None:
        update_changelog = guess_update_changelog(local_tree, subpath, cl)

    if update_changelog and only_changes_last_changelog_block(
            local_tree, changelog_path, changes):
        # If the script only changed the last entry in the changelog,
        # don't update the changelog
        update_changelog = False

    if update_changelog:
        add_changelog_entry(local_tree, changelog_path, summary)
        if specific_files:
            specific_files.append(changelog_path)

    description = result.description
    for tag in result.fixed_lintian_tags:
        description += "\n"
        description += "\n"
        description += "Fixes: lintian: %s\n" % tag
        description += (
            "See-also: https://lintian.debian.org/tags/%s.html\n" % tag)

    if committer is None:
        committer = get_committer(local_tree)

    revid = local_tree.commit(
        description, allow_pointless=False,
        reporter=NullCommitReporter(),
        committer=committer,
        specific_files=specific_files)
    result.revision_id = revid
    # TODO(jelmer): Run sbuild & verify lintian warning is gone?
    return result, summary


class ManyResult(object):

    def __init__(self):
        self.success = []
        self.failed_fixers = {}
        self.formatting_unpreservable = {}

    def __tuple__(self):
        return (self.success, self.failed_fixers)

    def __iter__(self):
        return iter(self.__tuple__())


def get_dirty_tracker(local_tree, subpath='', use_inotify=None):
    """Create a dirty tracker object."""
    if use_inotify is True:
        from .dirty_tracker import DirtyTracker
        return DirtyTracker(local_tree, subpath)
    elif use_inotify is False:
        return None
    else:
        try:
            from .dirty_tracker import DirtyTracker
        except ImportError:
            return None
        else:
            return DirtyTracker(local_tree, subpath)


def run_lintian_fixers(local_tree, fixers, update_changelog=True,
                       verbose=False, committer=None,
                       compat_release=None, minimum_certainty=None,
                       trust_package=False, allow_reformatting=False,
                       use_inotify=None, subpath='.', net_access=True,
                       opinionated=None, diligence=0):
    """Run a set of lintian fixers on a tree.

    Args:
      local_tree: WorkingTree object
      fixers: A set of Fixer objects
      update_changelog: Whether to add an entry to the changelog
      verbose: Whether to be verbose
      committer: Optional committer (name and email)
      compat_release: Minimum release that the package should be usable on
        (e.g. 'sid' or 'stretch')
      minimum_certainty: How certain the fixer should be
        about its changes.
      trust_package: Whether to run code from the package if necessary
      allow_reformatting: Whether to allow reformatting of changed files
      use_inotify: Use inotify to watch changes (significantly improves
        performance). Defaults to None (automatic)
      subpath: Subpath in the tree in which the package lives
      net_access: Whether to allow network access
      opinionated: Whether to be opinionated
      diligence: Level of diligence
    Returns:
      Tuple with two lists:
        1. list of tuples with (lintian-tag, certainty, description) of fixers
           that ran
        2. dictionary mapping fixer names for fixers that failed to run to the
           error that occurred
    """
    check_clean_tree(local_tree)
    fixers = list(fixers)
    dirty_tracker = get_dirty_tracker(
        local_tree, subpath=subpath, use_inotify=use_inotify)
    ret = ManyResult()
    with ui.ui_factory.nested_progress_bar() as pb:
        for i, fixer in enumerate(fixers):
            pb.update('Running fixer %r on %s' %
                      (fixer, local_tree.abspath(subpath)),
                      i, len(fixers))
            start = time.time()
            if dirty_tracker:
                dirty_tracker.mark_clean()
            try:
                result, summary = run_lintian_fixer(
                        local_tree, fixer, update_changelog=update_changelog,
                        committer=committer, compat_release=compat_release,
                        minimum_certainty=minimum_certainty,
                        trust_package=trust_package,
                        allow_reformatting=allow_reformatting,
                        dirty_tracker=dirty_tracker,
                        subpath=subpath, net_access=net_access,
                        opinionated=opinionated,
                        diligence=diligence)
            except FormattingUnpreservable as e:
                ret.formatting_unpreservable[fixer.name] = e.path
                if verbose:
                    note('Fixer %r was unable to preserve formatting of %s.',
                         fixer.name, e.path)
            except FixerFailed as e:
                ret.failed_fixers[fixer.name] = e
                if verbose:
                    note('Fixer %r failed to run.', fixer.name)
                    sys.stderr.write(str(e))
            except NotCertainEnough as e:
                if verbose:
                    note('Fixer %r made changes but not high enough '
                         'certainty (was %r, needed %r). (took: %.2fs)',
                         fixer.name, e.certainty, e.minimum_certainty,
                         time.time() - start)
            except NoChanges:
                if verbose:
                    note('Fixer %r made no changes. (took: %.2fs)',
                         fixer.name, time.time() - start)
            else:
                if verbose:
                    note('Fixer %r made changes. (took %.2fs)',
                         fixer.name, time.time() - start)
                ret.success.append((result, summary))
    return ret


def certainty_to_confidence(certainty):
    if certainty in ('unknown', None):
        return None
    return SUPPORTED_CERTAINTIES.index(certainty)


def confidence_to_certainty(confidence):
    if confidence is None:
        return 'unknown'
    try:
        return SUPPORTED_CERTAINTIES[confidence]
    except IndexError:
        raise ValueError(confidence)


def min_certainty(certainties):
    return confidence_to_certainty(
        max([certainty_to_confidence(c)
            for c in certainties] + [0]))



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Functions for working with watch files."""

from io import StringIO
import re
from warnings import warn

from lintian_brush.reformatting import (
    Updater,
    )


DEFAULT_VERSION = 4

# TODO(jelmer): Add support for substitution variables:

SUBSTITUTIONS = {
    # This is substituted with the source package name found in the first line
    # of the debian/changelog file.
    '@PACKAGE': None,
    # This is substituted by the legal upstream version regex (capturing).
    '@ANY_VERSION@': r'[-_]?(\d[\-+\.:\~\da-zA-Z]*)',
    # This is substituted by the typical archive file extension regex
    # (non-capturing).
    '@ARCHIVE_EXT@': r'(?i)\.(?:tar\.xz|tar\.bz2|tar\.gz|zip|tgz|tbz|txz)',
    # This is substituted by the typical signature file extension regex
    # (non-capturing).
    '@SIGNATURE_EXT@':
        r'(?i)\.(?:tar\.xz|tar\.bz2|tar\.gz|zip|tgz|tbz|txz)'
        r'\.(?:asc|pgp|gpg|sig|sign)',
    # This is substituted by the typical Debian extension regexp (capturing).
    '@DEB_EXT@': r'[\+~](debian|dfsg|ds|deb)(\.)?(\d+)?$',
}


class WatchFile(object):

    def __init__(self, entries=None, options=None, version=DEFAULT_VERSION):
        self.version = version
        if entries is None:
            entries = []
        self.entries = entries
        if options is None:
            options = []
        self.options = options

    def __iter__(self):
        return iter(self.entries)

    def dump(self, f):
        def serialize_options(opts):
            s = ','.join(opts)
            if ' ' in s or '\t' in s:
                return 'opts="' + s + '"'
            return 'opts=' + s
        f.write('version=%d\n' % self.version)
        if self.options:
            f.write(serialize_options(self.options) + '\n')
        for entry in self.entries:
            if entry.options:
                f.write(serialize_options(entry.options) + ' ')
            f.write(entry.url)
            if entry.matching_pattern:
                f.write(' ' + entry.matching_pattern)
            if entry.version:
                f.write(' ' + entry.version)
            if entry.script:
                f.write(' ' + entry.script)
            f.write('\n')


class Watch(object):

    def __init__(self, url, matching_pattern=None, version=None, script=None,
                 opts=None):
        self.url = url
        self.matching_pattern = matching_pattern
        self.version = version
        self.script = script
        if opts is None:
            opts = []
        self.options = opts

    def __repr__(self):
        return (
            "%s(%r, matching_pattern=%r, version=%r, script=%r, opts=%r)" % (
                self.__class__.__name__, self.url, self.matching_pattern,
                self.version, self.script, self.options))

    def __eq__(self, other):
        if not isinstance(other, Watch):
            return False
        return (other.url == self.url and
                other.matching_pattern == self.matching_pattern and
                other.version == self.version and
                other.script == self.script and
                other.options == self.options)

    def format_url(self, package):
        if '@PACKAGE@' not in self.url:
            return self.url
        if callable(package):
            package = package()
        return self.url.replace('@PACKAGE@', package)


class MissingVersion(Exception):
    """The version= line is missing."""


def parse_watch_file(f):
    """Parse a watch file.

    Args:
      f: watch file to parse
    """
    lines = []
    continued = []
    for line in f:
        if line.startswith('#'):
            continue
        if not line.strip():
            continue
        if line.rstrip('\n').endswith('\\'):
            continued.append(line.rstrip('\n\\'))
        else:
            continued.append(line)
            lines.append(continued)
            continued = []
    if continued:
        # Hmm, broken line?
        warn('watchfile ended with \\; skipping last line')
        lines.append([continued])
    if not lines:
        return None
    firstline = ''.join(lines.pop(0))
    try:
        key, value = firstline.split('=', 1)
    except ValueError:
        raise MissingVersion()
    if key.strip() != 'version':
        raise MissingVersion()
    version = int(value.strip())
    persistent_options = []
    entries = []
    for chunked in lines:
        if version > 3:
            chunked = [chunk.lstrip() for chunk in chunked]
        line = ''.join(chunked).strip()
        if not line:
            continue
        if line.startswith('opts='):
            if line[5] == '"':
                optend = line.index('"', 6)
                if optend == -1:
                    raise ValueError('Not matching " in %r' % line)
                opts = line[6:optend]
                line = line[optend+1:]
            else:
                try:
                    (opts, line) = line[5:].split(maxsplit=1)
                except ValueError:
                    opts = line[5:]
                    line = None
            opts = opts.split(',')
        else:
            opts = None
        if not line:
            persistent_options.extend(opts)
        else:
            try:
                url, line = line.split(maxsplit=1)
            except ValueError:
                url = line
                line = ''
            m = re.findall(r'/([^/]*\([^/]*\)[^/]*)$', url)
            if m:
                parts = [m[0]] + line.split(maxsplit=1)
                url = url[:-len(m[0])-1]
            else:
                parts = line.split(maxsplit=2)
            entries.append(Watch(url, *parts, opts=opts))
    return WatchFile(
        entries=entries, options=persistent_options, version=version)


class WatchUpdater(Updater):

    def __init__(self, path='debian/watch'):
        super(WatchUpdater, self).__init__(path)

    @property
    def watch_file(self):
        return self._parsed

    def _nonexistant(self):
        return None

    def _parse(self, content):
        return parse_watch_file(content.splitlines())

    def _format(self, parsed):
        nf = StringIO()
        parsed.dump(nf)
        return nf.getvalue()



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush."""

import os
import re
import sys

from debian.changelog import (
    Changelog,
    Version,
    )

from breezy.config import GlobalStack
from breezy.tests import (
    TestCase,
    TestCaseWithTransport,
    )

from lintian_brush import (
    Fixer,
    FixerFailed,
    FixerResult,
    FixerScriptFailed,
    NoChanges,
    NotDebianPackage,
    PendingChanges,
    PythonScriptFixer,
    ScriptFixer,
    UnsupportedCertainty,
    available_lintian_fixers,
    check_clean_tree,
    add_changelog_entry,
    guess_update_changelog,
    certainty_sufficient,
    min_certainty,
    certainty_to_confidence,
    confidence_to_certainty,
    get_committer,
    increment_version,
    only_changes_last_changelog_block,
    parse_script_fixer_output,
    run_lintian_fixer,
    run_lintian_fixers,
    select_fixers,
    version_string,
    )


CHANGELOG_FILE = ('debian/changelog', """\
blah (0.1) UNRELEASED; urgency=medium

  * Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""")


class AvailableLintianFixersTest(TestCaseWithTransport):

    def test_find_shell_scripts(self):
        self.build_tree([
            'fixers/',
            'fixers/anotherdir/',
            'fixers/i-fix-a-tag.sh',
            'fixers/i-fix-another-tag.py',
            'fixers/.hidden',
            'fixers/backup-file.sh~',
            'fixers/no-extension'])
        self.build_tree_contents([
            ('fixers/index.desc', """\
Fix-Script: foo.sh
Lintian-Tags: i-fix-a-tag

Fix-Script: bar.sh
Lintian-Tags: i-fix-another-tag, no-extension
""")])
        self.assertEqual(
                [['i-fix-a-tag'], ['i-fix-another-tag', 'no-extension']],
                [fixer.lintian_tags
                 for fixer in available_lintian_fixers('fixers')])


class CheckCleanTreeTests(TestCaseWithTransport):

    def make_test_tree(self, format=None):
        tree = self.make_branch_and_tree('.', format=format)
        self.build_tree_contents([
            ('debian/', ),
            ('debian/control', """\
Source: blah
Vcs-Git: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

"""),
            CHANGELOG_FILE])
        tree.add(['debian', 'debian/changelog', 'debian/control'])
        tree.commit('Initial thingy.')
        return tree

    def test_pending_changes(self):
        tree = self.make_test_tree()
        self.build_tree_contents([('debian/changelog', 'blah')])
        with tree.lock_write():
            self.assertRaises(
                PendingChanges, check_clean_tree, tree)

    def test_pending_changes_bzr_empty_dir(self):
        # See https://bugs.debian.org/914038
        tree = self.make_test_tree(format='bzr')
        self.build_tree_contents([('debian/upstream/', )])
        with tree.lock_write():
            self.assertRaises(
                PendingChanges, check_clean_tree, tree)

    def test_pending_changes_git_empty_dir(self):
        # See https://bugs.debian.org/914038
        tree = self.make_test_tree(format='git')
        self.build_tree_contents([('debian/upstream/', )])
        with tree.lock_write():
            check_clean_tree(tree)

    def test_pending_changes_git_dir_with_ignored(self):
        # See https://bugs.debian.org/914038
        tree = self.make_test_tree(format='git')
        self.build_tree_contents([
            ('debian/upstream/', ),
            ('debian/upstream/blah', ''),
            ('.gitignore', 'blah\n'),
            ])
        tree.add('.gitignore')
        tree.commit('add gitignore')
        with tree.lock_write():
            check_clean_tree(tree)

    def test_extra(self):
        tree = self.make_test_tree()
        self.build_tree_contents([('debian/foo', 'blah')])
        with tree.lock_write():
            self.assertRaises(
                PendingChanges, check_clean_tree,
                tree)


class DummyFixer(Fixer):

    def run(self, basedir, package, *args, **kwargs):
        with open(os.path.join(basedir, 'debian/control'), 'a') as f:
            f.write('a new line\n')
        return FixerResult("Fixed some tag.\nExtended description.",
                           ['some-tag'], "certain")


class FailingFixer(Fixer):

    def run(self, basedir, package, *args, **kwargs):
        with open(os.path.join(basedir, 'debian/foo'), 'w') as f:
            f.write("blah")
        with open(os.path.join(basedir, 'debian/control'), 'a') as f:
            f.write("foo\n")
        raise FixerFailed("Not successful")


class RunLintianFixerTests(TestCaseWithTransport):

    def make_test_tree(self, format=None):
        tree = self.make_branch_and_tree('.', format=format)
        self.build_tree_contents([
            ('debian/', ),
            ('debian/control', """\
Source: blah
Vcs-Git: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

"""),
            CHANGELOG_FILE])
        tree.add(['debian', 'debian/changelog', 'debian/control'])
        tree.commit('Initial thingy.')
        return tree

    def test_not_debian_tree(self):
        tree = self.make_test_tree()
        tree.remove('debian/changelog')
        os.remove('debian/changelog')
        tree.commit("not a debian dir")
        with tree.lock_write():
            self.assertRaises(
                NotDebianPackage, run_lintian_fixer,
                tree, DummyFixer('dummy', 'some-tag'),
                update_changelog=False)

    def test_simple_modify(self):
        tree = self.make_test_tree()
        with tree.lock_write():
            result, summary = run_lintian_fixer(
                tree, DummyFixer('dummy', 'some-tag'),
                update_changelog=False)
        self.assertEqual(summary, "Fixed some tag.")
        self.assertEqual(['some-tag'], result.fixed_lintian_tags)
        self.assertEqual('certain', result.certainty)
        self.assertEqual(2, tree.branch.revno())
        self.assertEqual(
                tree.get_file_lines('debian/control')[-1],
                b"a new line\n")

    def test_simple_modify_too_uncertain(self):
        tree = self.make_test_tree()

        class UncertainFixer(Fixer):
            def run(self, basedir, package, *args, **kwargs):
                with open(os.path.join(basedir, 'debian/somefile'), 'w') as f:
                    f.write("test")
                return FixerResult("Renamed a file.", certainty='possible')
        with tree.lock_write():
            self.assertRaises(
                NoChanges, run_lintian_fixer,
                tree, UncertainFixer('dummy', 'some-tag'),
                update_changelog=False, minimum_certainty='certain')
        self.assertEqual(1, tree.branch.revno())

    def test_simple_modify_acceptably_uncertain(self):
        tree = self.make_test_tree()

        class UncertainFixer(Fixer):
            def run(self, basedir, package, *args, **kwargs):
                with open(os.path.join(basedir, 'debian/somefile'), 'w') as f:
                    f.write("test")
                return FixerResult("Renamed a file.", certainty='possible')
        with tree.lock_write():
            result, summary = run_lintian_fixer(
                tree, UncertainFixer('dummy', 'some-tag'),
                update_changelog=False, minimum_certainty='possible')
        self.assertEqual(2, tree.branch.revno())

    def test_new_file(self):
        tree = self.make_test_tree()

        class NewFileFixer(Fixer):
            def run(self, basedir, package, *args, **kwargs):
                with open(os.path.join(basedir, 'debian/somefile'), 'w') as f:
                    f.write("test")
                return FixerResult("Created new file.", ['some-tag'])
        with tree.lock_write():
            result, summary = run_lintian_fixer(
                tree, NewFileFixer('new-file', 'some-tag'),
                update_changelog=False)
        self.assertEqual(summary, "Created new file.")
        self.assertIs(None, result.certainty)
        self.assertEqual(['some-tag'], result.fixed_lintian_tags)
        rev = tree.branch.repository.get_revision(
            tree.last_revision())
        self.assertEqual(rev.message, (
            'Created new file.\n'
            '\n'
            'Fixes: lintian: some-tag\n'
            'See-also: https://lintian.debian.org/tags/some-tag.html\n'))
        self.assertEqual(2, tree.branch.revno())
        basis_tree = tree.branch.basis_tree()
        with basis_tree.lock_read():
            self.assertEqual(
                    basis_tree.get_file_text('debian/somefile'),
                    b"test")

    def test_rename_file(self):
        tree = self.make_test_tree()

        class RenameFileFixer(Fixer):
            def run(self, basedir, package, *args, **kwargs):
                os.rename(os.path.join(basedir, 'debian/control'),
                          os.path.join(basedir, 'debian/control.blah'))
                return FixerResult("Renamed a file.")
        orig_basis_tree = tree.branch.basis_tree()
        with tree.lock_write():
            result, summary = run_lintian_fixer(
                tree, RenameFileFixer('rename', 'some-tag'),
                update_changelog=False)
        self.assertEqual(summary, "Renamed a file.")
        self.assertIs(None, result.certainty)
        self.assertEqual([], result.fixed_lintian_tags)
        self.assertEqual(2, tree.branch.revno())
        basis_tree = tree.branch.basis_tree()
        with basis_tree.lock_read(), orig_basis_tree.lock_read():
            self.assertFalse(basis_tree.has_filename('debian/control'))
            self.assertTrue(basis_tree.has_filename('debian/control.blah'))
            self.assertNotEqual(orig_basis_tree.get_revision_id(),
                                basis_tree.get_revision_id())
            self.expectFailure(
                'mv --auto appears to be broken',
                self.assertEqual, basis_tree.path2id('debian/control.blah'),
                orig_basis_tree.path2id('debian/control'))

    def test_empty_change(self):
        tree = self.make_test_tree()

        class EmptyFixer(Fixer):
            def run(self, basedir, package, *args, **kwargs):
                return FixerResult("I didn't actually change anything.")
        with tree.lock_write():
            self.assertRaises(
                    NoChanges, run_lintian_fixer, tree,
                    EmptyFixer('empty', 'some-tag'), update_changelog=False)
        self.assertEqual(1, tree.branch.revno())
        with tree.lock_read():
            self.assertEqual(
                [], list(tree.iter_changes(tree.basis_tree())))

    def test_fails(self):
        tree = self.make_test_tree()
        with tree.lock_write():
            self.assertRaises(
                    Exception, run_lintian_fixer, tree,
                    FailingFixer('fail', 'some-tag'), update_changelog=False)
        self.assertEqual(1, tree.branch.revno())
        with tree.lock_read():
            self.assertEqual(
                [], list(tree.iter_changes(tree.basis_tree())))


class RunLintianFixersTests(TestCaseWithTransport):

    def setUp(self):
        super(RunLintianFixersTests, self).setUp()
        self.tree = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/control', """\
Source: blah
Vcs-Git: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

"""),
            CHANGELOG_FILE])
        self.tree.add(['debian', 'debian/changelog', 'debian/control'])
        self.tree.commit('Initial thingy.')

    def test_fails(self):
        with self.tree.lock_write():
            applied, failed = run_lintian_fixers(
                    self.tree, [FailingFixer('fail', 'some-tag')],
                    update_changelog=False)
        self.assertEqual([], applied)
        self.assertEqual({'fail': FixerFailed('Not successful')}, failed)
        with self.tree.lock_read():
            self.assertEqual(
                [], list(self.tree.iter_changes(self.tree.basis_tree())))

    def test_not_debian_tree(self):
        self.tree.remove('debian/changelog')
        os.remove('debian/changelog')
        self.tree.commit("not a debian dir")
        with self.tree.lock_write():
            self.assertRaises(
                NotDebianPackage, run_lintian_fixers,
                self.tree, [DummyFixer('dummy', 'some-tag')],
                update_changelog=False)

    def test_simple_modify(self):
        with self.tree.lock_write():
            applied, failed = run_lintian_fixers(
                self.tree, [DummyFixer('dummy', 'some-tag')],
                update_changelog=False)
            revid = self.tree.last_revision()
        self.assertEqual(
                [(FixerResult("Fixed some tag.\nExtended description.",
                              ['some-tag'], 'certain', revision_id=revid),
                  'Fixed some tag.')],
                applied)
        self.assertEqual({}, failed)
        self.assertEqual(2, self.tree.branch.revno())
        self.assertEqual(
                self.tree.get_file_lines('debian/control')[-1],
                b"a new line\n")


class HonorsVcsCommitter(TestCaseWithTransport):

    def make_package_tree(self, format):
        tree = self.make_branch_and_tree('.', format=format)
        self.build_tree_contents([
            ('debian/', ),
            ('debian/control', """\
Source: blah
Vcs-Git: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

"""),
            CHANGELOG_FILE])
        tree.add(['debian', 'debian/changelog', 'debian/control'])
        tree.commit('Initial thingy.')
        return tree

    def make_change(self, tree, committer=None):
        with tree.lock_write():
            result, summary = run_lintian_fixer(
                tree, DummyFixer('dummy', 'some-tag'),
                update_changelog=False, committer=committer)
        self.assertEqual(summary, "Fixed some tag.")
        self.assertEqual(['some-tag'], result.fixed_lintian_tags)
        self.assertEqual('certain', result.certainty)
        self.assertEqual(2, tree.branch.revno())
        self.assertEqual(
                tree.get_file_lines('debian/control')[-1],
                b"a new line\n")

    def test_honors_tree_committer_specified(self):
        tree = self.make_package_tree('git')
        self.make_change(tree, committer='Jane Example <jane@example.com>')
        rev = tree.branch.repository.get_revision(tree.branch.last_revision())
        self.assertEqual(rev.committer, 'Jane Example <jane@example.com>')

    def test_honors_tree_committer_config(self):
        tree = self.make_package_tree('git')
        with open(os.path.join(tree.basedir, '.git/config'), 'w') as f:
            f.write("""\
[user]
  email = jane@example.com
  name = Jane Example

""")

        self.make_change(tree)
        rev = tree.branch.repository.get_revision(tree.branch.last_revision())
        self.assertEqual(rev.committer, 'Jane Example <jane@example.com>')


class IncrementVersionTests(TestCase):

    def assertVersion(self, expected, start):
        v = Version(start)
        increment_version(v)
        self.assertEqual(Version(expected), v)

    def test_full(self):
        self.assertVersion('1.0-2', '1.0-1')

    def test_native(self):
        self.assertVersion('1.1', '1.0')


class OnlyChangesLastChangelogBlockTests(TestCaseWithTransport):

    def make_package_tree(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/control', """\
Source: blah
Vcs-Git: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

"""),
            ('debian/changelog', """\
blah (0.2) UNRELEASED; urgency=medium

  * And a change.

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100

blah (0.1) unstable; urgency=medium

  * Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""")])
        tree.add(['debian', 'debian/changelog', 'debian/control'])
        tree.commit('Initial thingy.')
        return tree

    def test_no_changes(self):
        tree = self.make_package_tree()
        basis_tree = tree.basis_tree()
        with tree.lock_read(), basis_tree.lock_read():
            changes = tree.iter_changes(basis_tree)
        self.assertFalse(only_changes_last_changelog_block(
            tree, 'debian/changelog', changes))

    def test_other_change(self):
        tree = self.make_package_tree()
        self.build_tree_contents([
            ('debian/control', """\
Source: blah
Vcs-Bzr: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

""")])
        basis_tree = tree.basis_tree()
        with tree.lock_read(), basis_tree.lock_read():
            changes = tree.iter_changes(basis_tree)
        self.assertFalse(only_changes_last_changelog_block(
            tree, 'debian/changelog', changes))

    def test_other_changes(self):
        tree = self.make_package_tree()
        self.build_tree_contents([
            ('debian/control', """\
Source: blah
Vcs-Bzr: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

"""),
            ('debian/changelog', """\
blah (0.1) UNRELEASED; urgency=medium

  * Initial release. (Closes: #911016)
  * Some other change.

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""")])
        basis_tree = tree.basis_tree()
        with tree.lock_read(), basis_tree.lock_read():
            changes = tree.iter_changes(basis_tree)
        self.assertFalse(only_changes_last_changelog_block(
            tree, 'debian/changelog', changes))

    def test_changes_to_other_changelog_entries(self):
        tree = self.make_package_tree()
        self.build_tree_contents([
            ('debian/changelog', """\
blah (0.2) UNRELEASED; urgency=medium

  * debian/changelog: And a change.

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100

blah (0.1) unstable; urgency=medium

  * debian/changelog: Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""")])
        basis_tree = tree.basis_tree()
        with tree.lock_read(), basis_tree.lock_read():
            changes = tree.iter_changes(basis_tree)
        self.assertFalse(only_changes_last_changelog_block(
            tree, 'debian/changelog', changes))

    def test_changes_to_last_only(self):
        tree = self.make_package_tree()
        self.build_tree_contents([
            ('debian/changelog', """\
blah (0.2) UNRELEASED; urgency=medium

  * And a change.
  * Not a team upload.

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100

blah (0.1) unstable; urgency=medium

  * Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""")])
        basis_tree = tree.basis_tree()
        with tree.lock_read(), basis_tree.lock_read():
            changes = tree.iter_changes(basis_tree)
        self.assertTrue(only_changes_last_changelog_block(
            tree, 'debian/changelog', changes))

    def test_changes_to_last_only_but_released(self):
        tree = self.make_package_tree()
        self.build_tree_contents([
            ('debian/changelog', """\
blah (0.2) unstable; urgency=medium

  * And a change.

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100

blah (0.1) unstable; urgency=medium

  * Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""")])
        tree.commit('release')
        self.build_tree_contents([
            ('debian/changelog', """\
blah (0.2) unstable; urgency=medium

  * And a change.
  * Team Upload.

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100

blah (0.1) unstable; urgency=medium

  * Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""")])
        basis_tree = tree.basis_tree()
        with tree.lock_read(), basis_tree.lock_read():
            changes = tree.iter_changes(basis_tree)
        self.assertFalse(only_changes_last_changelog_block(
            tree, 'debian/changelog', changes))


class LintianBrushVersion(TestCase):

    def test_matches_package_version(self):
        self.skipTest('dch on Ubuntu delta.')
        return
        if not os.path.exists('debian/changelog'):
            self.skipTest(
                'no debian/changelog available. '
                'Running outside of source tree?')
        with open('debian/changelog', 'r') as f:
            cl = Changelog(f, max_blocks=1)
        self.assertEqual(str(cl.version).split('+')[0], version_string)

    def test_matches_setup_version(self):
        if not os.path.exists('setup.py'):
            self.skipTest(
                'no setup.py available. '
                'Running outside of source tree?')
        # TODO(jelmer): Surely there's a better way of doing this?
        with open('setup.py', 'r') as f:
            for l in f:
                m = re.match(r'[ ]*version="(.*)",', l)
                if m:
                    setup_version = m.group(1)
                    break
            else:
                raise AssertionError('setup version not found')
        self.assertEqual(version_string, setup_version)


class GetCommitterTests(TestCaseWithTransport):

    def test_git_env(self):
        self.overrideEnv('GIT_COMMITTER_NAME', 'Some Git Committer')
        self.overrideEnv('GIT_COMMITTER_EMAIL', 'committer@example.com')
        tree = self.make_branch_and_tree('.', format='git')
        self.assertEqual(
            'Some Git Committer <committer@example.com>',
            get_committer(tree))

    def test_git_config(self):
        tree = self.make_branch_and_tree('.', format='git')
        with open('.git/config', 'w') as f:
            f.write("""\
[user]
   name = Some Other Git Committer
   email = other@example.com
""")
        self.assertEqual(
            'Some Other Git Committer <other@example.com>',
            get_committer(tree))

    def test_global_stack(self):
        gs = GlobalStack()
        gs.set('email', 'Yet Another Committer <yet@example.com>')
        tree = self.make_branch_and_tree('.', format='git')
        self.assertEqual(
            'Yet Another Committer <yet@example.com>',
            get_committer(tree))


class CertaintySufficientTests(TestCase):

    def test_sufficient(self):
        self.assertTrue(certainty_sufficient('certain', 'certain'))
        self.assertTrue(certainty_sufficient('certain', 'possible'))
        self.assertTrue(certainty_sufficient('certain', None))
        self.assertTrue(certainty_sufficient('possible', None))
        # TODO(jelmer): Should we really always allow unknown certainties
        # through?
        self.assertTrue(certainty_sufficient(None, 'certain'))

    def test_insufficient(self):
        self.assertFalse(certainty_sufficient('possible', 'certain'))


class CertaintyVsConfidenceTests(TestCase):

    def test_confidence_to_certainty(self):
        self.assertEqual("certain", confidence_to_certainty(0))
        self.assertEqual("confident", confidence_to_certainty(1))
        self.assertEqual("likely", confidence_to_certainty(2))
        self.assertEqual("possible", confidence_to_certainty(3))
        self.assertEqual("unknown", confidence_to_certainty(None))
        self.assertRaises(ValueError, confidence_to_certainty, 2000)

    def test_certainty_to_confidence(self):
        self.assertEqual(0, certainty_to_confidence("certain"))
        self.assertEqual(1, certainty_to_confidence("confident"))
        self.assertEqual(2, certainty_to_confidence("likely"))
        self.assertEqual(3, certainty_to_confidence("possible"))
        self.assertIs(None, certainty_to_confidence("unknown"))
        self.assertRaises(ValueError, certainty_to_confidence, "blah")


class MinimumCertaintyTests(TestCase):

    def test_minimum(self):
        self.assertEqual("certain", min_certainty([]))
        self.assertEqual("certain", min_certainty(["certain"]))
        self.assertEqual("possible", min_certainty(["possible"]))
        self.assertEqual("possible", min_certainty(["possible", "certain"]))
        self.assertEqual("likely", min_certainty(["likely", "certain"]))
        self.assertEqual(
            "possible",
            min_certainty(["likely", "certain", "possible"]))


class ParseScriptFixerOutputTests(TestCase):

    def test_simple(self):
        self.assertEqual(
            FixerResult('Do bla', ['tag-a'], 'certain'),
            parse_script_fixer_output("""\
Do bla
Fixed-Lintian-Tags: tag-a
Certainty: certain
"""))

    def test_unknown_certainty(self):
        self.assertRaises(
            UnsupportedCertainty,
            parse_script_fixer_output, """\
Do bla
Fixed-Lintian-Tags: tag-a
Certainty: blah
""")

    def test_default_certainty(self):
        self.assertEqual(
            FixerResult('Do bla', ['tag-a'], None),
            parse_script_fixer_output("""\
Do bla
Fixed-Lintian-Tags: tag-a
"""))

    def test_patch_name(self):
        self.assertEqual(
            FixerResult('Do bla', ['tag-a'], 'certain', 'aname'),
            parse_script_fixer_output("""\
Do bla
Fixed-Lintian-Tags: tag-a
Certainty: certain
Patch-Name: aname
"""))


class BaseScriptFixerTests(object):

    script_fixer_cls = None

    def create_fixer(self, code):
        self.build_tree_contents([
            ('script.py',
             "#!" + sys.executable + "\n" + code)])
        os.chmod('script.py', 0o755)
        fixer = self.script_fixer_cls(
            'fixer', ['a-tag'], os.path.abspath('script.py'))
        self.assertEqual(os.path.abspath('script.py'), fixer.script_path)
        return fixer

    def test_noop(self):
        fixer = self.create_fixer("print('I did not do anything')\n")
        result = fixer.run(self.test_dir, 'blah', '0.1', 'buster')
        self.assertIsInstance(result, FixerResult)
        self.assertEqual(result.description, 'I did not do anything')

    def test_chdir(self):
        fixer = self.create_fixer("import os; print(os.getcwd())\n")
        os.mkdir('subdir')
        os.chdir('subdir')
        result = fixer.run(self.test_dir, 'blah', '0.1', 'buster')
        self.assertIsInstance(result, FixerResult)
        self.assertEqual(result.description, self.test_dir)

    def test_exception(self):
        fixer = self.create_fixer("""\
def foo():
    raise Exception('oops')
foo()
""")
        e = self.assertRaises(
            FixerScriptFailed, fixer.run, self.test_dir, 'blah', '0.1',
            'buster')
        self.assertEqual(e.path, fixer.script_path)
        self.assertEqual(e.returncode, 1)
        self.assertEqual(
            e.errors.splitlines()[-2:],
            ["    raise Exception('oops')", 'Exception: oops'])


class ScriptFixerTests(BaseScriptFixerTests, TestCaseWithTransport):

    script_fixer_cls = ScriptFixer


class PythonScriptFixerTests(BaseScriptFixerTests, TestCaseWithTransport):

    script_fixer_cls = PythonScriptFixer


class SelectFixersTests(TestCase):

    def test_exists(self):
        self.assertEqual(
            ['dummy1'],
            [f.name for f in select_fixers(
                [DummyFixer('dummy1', 'some-tag'),
                 DummyFixer('dummy2', 'other-tag')],
                ['dummy1'])])

    def test_missing(self):
        self.assertRaises(
            KeyError, select_fixers, [DummyFixer('dummy', 'some-tag')],
            ['other'])

    def test_exclude_missing(self):
        self.assertRaises(
            KeyError, select_fixers, [DummyFixer('dummy', 'some-tag')],
            ['dummy'], ['some-other'])

    def test_exclude(self):
        self.assertEqual(
            ['dummy1'],
            [f.name for f in select_fixers(
                [DummyFixer('dummy1', 'some-tag'),
                 DummyFixer('dummy2', 'other-tag')],
                ['dummy1', 'dummy2'], ['dummy2'])])


class ChangelogAddEntryTests(TestCaseWithTransport):

    def test_edit_existing(self):
        self.skipTest('dch on Ubuntu delta.')
        return
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/changelog', """\
lintian-brush (0.35) UNRELEASED; urgency=medium

  * Support updating templated debian/control files that use cdbs
    template.

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
 """)])
        tree.add(['debian', 'debian/changelog'])
        self.overrideEnv('DEBFULLNAME', 'Jane Example')
        self.overrideEnv('DEBEMAIL', 'jane@example.com')
        add_changelog_entry(tree, 'debian/changelog', 'Add a foo')
        self.assertFileEqual("""\
lintian-brush (0.35) UNRELEASED; urgency=medium

  [ Joe Example ]
  * Support updating templated debian/control files that use cdbs
    template.

  [ Jane Example ]
  * Add a foo

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""", 'debian/changelog')

    def test_add_new(self):
        self.skipTest('dch on Ubuntu delta.')
        return
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/changelog', """\
lintian-brush (0.35) unstable; urgency=medium

  * Support updating templated debian/control files that use cdbs
    template.

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
 """)])
        tree.add(['debian', 'debian/changelog'])
        self.overrideEnv('DEBFULLNAME', 'Jane Example')
        self.overrideEnv('DEBEMAIL', 'jane@example.com')
        add_changelog_entry(tree, 'debian/changelog', 'Add a foo')
        text = tree.get_file_text('debian/changelog').decode()
        lines = text.splitlines()
        self.assertEqual([
            'lintian-brush (0.36) UNRELEASED; urgency=medium',
            '',
            '  * Add a foo',
            ''],
            lines[:4])
        self.assertTrue(
            lines[4].startswith(' -- Jane Example <jane@example.com>  '),
            lines[4])

    def test_edit_broken_first_line(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/changelog', """\
THIS IS NOT A PARSEABLE LINE
lintian-brush (0.35) UNRELEASED; urgency=medium

  * Support updating templated debian/control files that use cdbs
    template.

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
 """)])
        tree.add(['debian', 'debian/changelog'])
        self.overrideEnv('DEBFULLNAME', 'Jane Example')
        self.overrideEnv('DEBEMAIL', 'jane@example.com')
        add_changelog_entry(tree, 'debian/changelog', 'Add a foo')
        self.assertFileEqual("""\
THIS IS NOT A PARSEABLE LINE
lintian-brush (0.35) UNRELEASED; urgency=medium

  [ Joe Example ]
  * Support updating templated debian/control files that use cdbs
    template.

  [ Jane Example ]
  * Add a foo

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""", 'debian/changelog')


class GuessUpdateChangelogTests(TestCaseWithTransport):

    def test_no_gbp_conf(self):
        tree = self.make_branch_and_tree('.')
        self.assertTrue(guess_update_changelog(tree))

    def test_gbp_conf_dch(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/gbp.conf', """\
[dch]
""")])
        tree.add(['debian', 'debian/gbp.conf'])
        self.assertFalse(guess_update_changelog(tree))

    def test_changelog_sha_prefixed(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/changelog', """\
blah (0.20.1) unstable; urgency=medium

  [ Somebody ]
  * [ebb7c31] do a thing
  * [629746a] do another thing that actually requires us to wrap lines
    and then

  [ Somebody Else ]
  * [b02b435] do another thing

 -- Joe User <joe@example.com>  Tue, 19 Nov 2019 15:29:47 +0100
""")])
        tree.add(['debian', 'debian/changelog'])
        self.assertFalse(guess_update_changelog(tree))



#!/usr/bin/python3
# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with licenses."""

import json
import os

COMMON_LICENSES_DIR = '/usr/share/common-licenses'

FULL_LICENSE_NAME = {
    'Apache-2.0': 'Apache License, Version 2.0',
    'CC0-1.0': 'CC0 1.0 Universal license',
}


def load_spdx_data():
    path = os.path.abspath(os.path.join(
        os.path.dirname(__file__), '..', 'spdx.json'))
    if not os.path.isfile(path):
        import pkg_resources
        path = pkg_resources.resource_filename(
            __name__, 'lintian-brush/spdx.json')
        if not os.path.isfile(path):
            # Urgh.
            path = '/usr/share/lintian-brush/spdx.json'
    with open(path, 'rb') as f:
        return json.load(f)



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.salsa."""

from breezy.tests import (
    TestCase,
    )
from lintian_brush.salsa import (
    guess_repository_url,
    determine_browser_url,
    salsa_url_from_alioth_url,
    )


class GuessRepositoryURLTests(TestCase):

    def test_unknown(self):
        self.assertIs(
            None,
            guess_repository_url(
                'blah', 'unknown-team@lists.alioth.debian.org'))

    def test_individual(self):
        self.assertEqual(
            'https://salsa.debian.org/jelmer/lintian-brush.git',
            guess_repository_url('lintian-brush', 'jelmer@debian.org'))

    def test_team(self):
        self.assertEqual(
            'https://salsa.debian.org/js-team/node-blah.git',
            guess_repository_url(
                'node-blah', 'pkg-javascript-devel@lists.alioth.debian.org'))


class DetermineBrowserUrlTests(TestCase):

    def test_browser_url(self):
        self.assertEqual(
            'https://salsa.debian.org/js-team/node-blah',
            determine_browser_url(
                'https://salsa.debian.org/js-team/node-blah.git'))
        self.assertEqual(
            'https://salsa.debian.org/js-team/node-blah',
            determine_browser_url(
                'https://salsa.debian.org/js-team/node-blah'))

    def test_branch(self):
        self.assertEqual(
            'https://salsa.debian.org/js-team/node-blah/tree/debian',
            determine_browser_url(
                'https://salsa.debian.org/js-team/node-blah -b debian'))

    def test_subpath(self):
        self.assertEqual(
            'https://salsa.debian.org/js-team/node-blah/tree/debian/foo',
            determine_browser_url(
                'https://salsa.debian.org/js-team/node-blah -b debian [foo]'))
        self.assertEqual(
            'https://salsa.debian.org/js-team/node-blah/tree/HEAD/foo',
            determine_browser_url(
                'https://salsa.debian.org/js-team/node-blah [foo]'))


class SalsaUrlFromAliothUrlTests(TestCase):

    def test_none(self):
        self.assertIs(None, salsa_url_from_alioth_url(None, None))

    def test_mismatch(self):
        self.assertIs(
            None, salsa_url_from_alioth_url(
                'bzr', 'https://code.launchpad.net/blah'))

    def test_perl(self):
        self.assertEqual(
            'https://salsa.debian.org/perl-team/modules/packages/libbla.git',
            salsa_url_from_alioth_url(
                'svn', 'svn://svn.debian.org/pkg-perl/trunk/libbla'))

    def test_git(self):
        self.assertEqual(
            'https://salsa.debian.org/jelmer/python-bla.git',
            salsa_url_from_alioth_url(
                'git',
                'http://anonscm.debian.org/git/users/jelmer/python-bla')
            )
        self.assertEqual(
            'https://salsa.debian.org/jelmer/python-bla.git',
            salsa_url_from_alioth_url(
                'git',
                'http://anonscm.debian.org/users/jelmer/python-bla')
            )
        self.assertEqual(
            'https://salsa.debian.org/go-team/golang-example-blah-blah.git',
            salsa_url_from_alioth_url(
                'git',
                'http://anonscm.debian.org/pkg-go/golang-example-blah-blah')
            )
        self.assertEqual(
            'https://salsa.debian.org/lua-team/blah.git',
            salsa_url_from_alioth_url(
                'git',
                'https://alioth.debian.org/anonscm/git/pkg-lua/blah'))
        self.assertEqual(
            'https://salsa.debian.org/science-team/ros-geometry.git',
            salsa_url_from_alioth_url(
                'git',
                'https://anonscm.debian.org/cgit/debian-science/'
                'ros-geometry.git'))
        self.assertEqual(
            'https://salsa.debian.org/nagios-team/pkg-check-multi.git',
            salsa_url_from_alioth_url(
                'git',
                'https://anonscm.debian.org/git/pkg-nagios/'
                'pkg-check-multi.git'))
        self.assertEqual(
            'https://salsa.debian.org/perl-team/modules/packages/'
            'libgstream-interfaces-perl.git',
            salsa_url_from_alioth_url(
                'git',
                'git://git.debian.org/pkg-perl/packages/'
                'libgstream-interfaces-perl.git'))
        # TODO(jelmer): This should actually be
        # https://salsa.debian.org/qt-kde-team/extras/plasma-widget-menubar.git
        self.assertEqual(
            'https://salsa.debian.org/qt-kde-team/'
            'kde-extras/plasma-widget-menubar.git',
            salsa_url_from_alioth_url(
                'git',
                'git://anonscm.debian.org/pkg-kde/kde-extras/'
                'plasma-widget-menubar.git'))
        self.assertEqual(
            'https://salsa.debian.org/fonts-team/fonts-beteckna.git',
            salsa_url_from_alioth_url(
                'git',
                'git://anonscm.debian.org/pkg-fonts/fonts-beteckna.git'))
        self.assertEqual(
            'https://salsa.debian.org/brlink/xwit.git',
            salsa_url_from_alioth_url(
                'git',
                'git://anonscm.debian.org/users/brlink/xwit.git'))
        self.assertEqual(
            'https://salsa.debian.org/qt-kde-team/kde/kruler.git',
            salsa_url_from_alioth_url(
                'git',
                'https://anonscm.debian.org/git/pkg-kde/applications/'
                'kruler'))
        self.assertEqual(
            'https://salsa.debian.org/3dprinting-team/arduino-mighty.git',
            salsa_url_from_alioth_url(
                'git',
                'https://anonscm.debian.org/git/3dprinter/packages/'
                'arduino-mighty'))
        self.assertEqual(
            'https://salsa.debian.org/emacsen-team/lua-mode.git',
            salsa_url_from_alioth_url(
                'git',
                'https://anonscm.debian.org/git/pkg-emacsen/pkg/lua-mode'))
        self.assertEqual(
            'https://salsa.debian.org/debian-astro-team/astromatic.git',
            salsa_url_from_alioth_url(
                'git',
                'http://anonscm.debian.org/cgit/debian-astro/packages/'
                'astromatic.git'))
        self.assertEqual(
            'https://salsa.debian.org/debichem-team/bkchem.git',
            salsa_url_from_alioth_url(
                'git',
                'https://anonscm.debian.org/git/debichem/packages/bkchem.git'))
        self.assertEqual(
            'https://salsa.debian.org/3dprinting-team/yagv.git',
            salsa_url_from_alioth_url(
                'git',
                'https://anonscm.debian.org/cgit/3dprinter/packages/yagv.git'))

    def test_svn(self):
        self.assertEqual(
            'https://salsa.debian.org/multimedia-team/ezstream.git',
            salsa_url_from_alioth_url(
                'svn',
                'svn://svn.debian.org/svn/pkg-icecast/ezstream/trunk')
            )
        self.assertEqual(
            'https://salsa.debian.org/bsd-team/freebsd-buildutils.git',
            salsa_url_from_alioth_url(
                'svn',
                'svn://anonscm.debian.org/glibc-bsd/trunk/freebsd-buildutils/')
            )
        self.assertEqual(
            'https://salsa.debian.org/nvidia-team/nvclock.git',
            salsa_url_from_alioth_url(
                'svn',
                'svn://svn.debian.org/pkg-nvidia/packages/nvclock/trunk'
            ))
        self.assertEqual(
            'https://salsa.debian.org/llvm-team/llvm.git',
            salsa_url_from_alioth_url(
                'svn',
                'svn://svn.debian.org/svn/pkg-llvm/llvm/trunk/'
            ))
        self.assertEqual(
            'https://salsa.debian.org/xfce-team/xfswitch-plugin.git',
            salsa_url_from_alioth_url(
                'svn',
                'svn://anonscm.debian.org/pkg-xfce/goodies/trunk/'
                'xfswitch-plugin/'
            ))
        self.assertEqual(
            'https://salsa.debian.org/python-team/applications/pypar2.git',
            salsa_url_from_alioth_url(
                'svn',
                'svn://anonscm.debian.org/python-apps/packages/pypar2/trunk/'
            ))
        self.assertEqual(
            'https://salsa.debian.org/xml-sgml-team/docbook-slides-demo.git',
            salsa_url_from_alioth_url(
                'svn',
                'svn://svn.debian.org/debian-xml-sgml/packages/'
                'docbook-slides-demo/trunk/'
            ))
        self.assertEqual(
            'https://salsa.debian.org/fonts-team/ttf-inconsolata.git',
            salsa_url_from_alioth_url(
                'svn',
                'svn://svn.debian.org/svn/pkg-fonts/packages/ttf-inconsolata'
            ))
        self.assertEqual(
            'https://salsa.debian.org/qt-kde-team/kde3libs.git',
            salsa_url_from_alioth_url(
                'svn',
                'svn://svn.debian.org/svn/pkg-kde/trunk/packages/kde3libs'
            ))
        self.assertEqual(
            'https://salsa.debian.org/python-team/applications/'
            'upnp-inspector.git',
            salsa_url_from_alioth_url(
                'svn',
                'svn://svn.debian.org/python-apps/packages/upnp-inspector/'
                'trunk/'
            ))
        self.assertEqual(
            'https://salsa.debian.org/python-team/applications/hotssh.git',
            salsa_url_from_alioth_url(
                'svn',
                'svn://anonscm.debian.org/python-apps/packages/hotssh/trunk/'
            ))
        self.assertEqual(
            'https://salsa.debian.org/debichem-team/drawxtl.git',
            salsa_url_from_alioth_url(
                'svn',
                'svn://svn.debian.org/svn/debichem/unstable/drawxtl'))



#!/usr/bin/python3

from lintian_brush.control import (
    update_control,
    )

from debian.changelog import Changelog
from email.utils import parseaddr
import sys

TEAM_UPLOAD_LINE = '  * Team upload.'
uploader_emails = []


def check_uploaders(control):
    for entry in control.get('Uploaders', '').split(','):
        uploader_emails.append(parseaddr(entry)[1])


update_control(source_package_cb=check_uploaders)

with open('debian/changelog', 'r') as f:
    cl = Changelog(f.read())

last_change = cl[0]
has_team_upload = (TEAM_UPLOAD_LINE in last_change.changes())
name, email = parseaddr(last_change.author)
if email not in uploader_emails or not has_team_upload:
    sys.exit(2)

i = last_change._changes.index(TEAM_UPLOAD_LINE)
del last_change._changes[i]
if i > 0 and last_change._changes[i-1] == '' and last_change._changes[i] == '':
    # Also remove the next line, if it's empty
    del last_change._changes[i]

with open('debian/changelog', 'w') as f:
    f.write(str(cl))

print("Remove unnecessary Team Upload line in changelog.")
print("Fixed-Lintian-Tags: unnecessary-team-upload")



#!/usr/bin/python3
import json
import os
import re
import subprocess
import sys
from urllib.request import urlopen, Request

from lintian_brush import (
    USER_AGENT,
    DEFAULT_URLLIB_TIMEOUT,
    certainty_to_confidence,
    )
from lintian_brush.fixer import net_access_allowed
from lintian_brush.watch import WatchFile, Watch


if os.path.exists('debian/watch'):
    # Nothing to do here..
    sys.exit(0)

candidates = []

if os.path.exists('setup.py'):
    certainty = 'likely'
    try:
        lines = subprocess.check_output(
            ['python3', 'setup.py', '--name', '--version']).splitlines()
    except subprocess.CalledProcessError:
        # Worth a shot..
        lines = subprocess.check_output(
            ['python2', 'setup.py', '--name', '--version']).splitlines()
    lines = [line for line in lines if not line.startswith(b'W: ')]
    (project, version) = lines
    project = project.decode()
    version = version.decode()
    current_version_filenames = None
    if net_access_allowed():
        json_url = 'https://pypi.python.org/pypi/%s/json' % project
        headers = {'User-Agent': USER_AGENT}
        response = urlopen(
            Request(json_url, headers=headers), timeout=DEFAULT_URLLIB_TIMEOUT)
        pypi_data = json.load(response)
        if version in pypi_data['releases']:
            release = pypi_data['releases'][version]
            current_version_filenames = [
                (d['filename'], d['has_sig'])
                for d in release if d['packagetype'] == 'sdist']
    filename_regex = (
        r'%(project)s-(.+)\.(?:zip|tgz|tbz|txz|(?:tar\.(?:gz|bz2|xz)))' % {
            'project': project})
    opts = []
    # TODO(jelmer): Set uversionmangle?
    # opts.append('uversionmangle=s/(rc|a|b|c)/~$1/')
    if current_version_filenames:
        for (fn, has_sig) in current_version_filenames:
            if re.match(filename_regex, fn):
                certainty = 'certain'
                if has_sig:
                    opts.append('pgpsigurlmangle=s/$/.asc/')
    url = (r'https://pypi.debian.net/%(project)s/%(fname_regex)s' % {
        'project': project, 'fname_regex': filename_regex})
    # TODO(jelmer): Add pgpsigurlmangle if has_sig==True
    w = Watch(url, opts=opts)
    candidates.append((w, 'pypi', certainty))


def guess_github_watch_entry(parsed_url, upstream_version):
    from breezy.branch import Branch
    import re
    if not net_access_allowed():
        return None, None
    branch = Branch.open(code['Repository'])
    tags = branch.tags.get_tag_dict()
    POSSIBLE_PATTERNS = [r"v(\d\S+)", r"(\d\S+)"]
    version_pattern = None
    # TODO(jelmer): Maybe use releases API instead?
    # TODO(jelmer): Automatically added mangling for
    # e.g. rc and beta
    uversionmangle = []
    for name in sorted(tags, reverse=True):
        for pattern in POSSIBLE_PATTERNS:
            m = re.match(pattern, name)
            if not m:
                continue
            if m.group(1) == upstream_version:
                version_pattern = pattern
                break
        if version_pattern:
            break
    else:
        return None, None
    (username, project) = parsed_url.path.strip('/').split('/')
    if project.endswith('.git'):
        project = project[:-4]
    download_url = 'https://github.com/%(user)s/%(project)s/tags' % {
        'user': username, 'project': project}
    matching_pattern = r'.*\/%s\.tar\.gz' % version_pattern
    opts = []
    opts.append(
        r'filenamemangle=s/%(pattern)s/%(project)s-$1\.tar\.gz/' % {
            'pattern': matching_pattern,
            'project': project})
    if uversionmangle:
        opts.append(r'uversionmangle=' + ';'.join(uversionmangle))
    # TODO(jelmer): Check for GPG
    # opts.append(
    #    r'pgpsigurlmangle='
    #    r's/archive\/%s\.tar\.gz/releases\/download\/$1\/$1\.tar\.gz\.asc/' %
    #    version_pattern)
    return Watch(download_url, matching_pattern, opts=opts), 'certain'


try:
    with open('debian/upstream/metadata', 'r') as f:
        inp = f.read()
except FileNotFoundError:
    pass
else:
    import ruamel.yaml
    code = ruamel.yaml.round_trip_load(inp, preserve_quotes=True)

    from urllib.parse import urlparse
    from debian.changelog import Version

    try:
        parsed_url = urlparse(code['Repository'])
    except KeyError:
        pass
    else:
        upstream_version = Version(
            os.environ['CURRENT_VERSION']).upstream_version
        if parsed_url.hostname == 'github.com':
            w, certainty = guess_github_watch_entry(
                parsed_url, upstream_version)
            if w:
                candidates.append((w, 'github', certainty))


if not candidates:
    sys.exit(0)

winner = None
for candidate in candidates:
    if winner is not None and (
            certainty_to_confidence(candidate[2]) >=
            certainty_to_confidence(winner[2])):
        continue
    winner = candidate

wf = WatchFile()
(entry, site, certainty) = winner
wf.entries.append(winner[0])

# TODO(jelmer): Find candidate with highest certainty
with open('debian/watch', 'w') as f:
    wf.dump(f)

print("Add debian/watch file, using %s." % site)
print("Certainty: %s" % certainty)
print("Fixed-Lintian-Tags: debian-watch-file-is-missing")



#!/usr/bin/python3

from lintian_brush.changelog import ChangelogUpdater, changes_by_author
from lintian_brush.fixer import report_result

with ChangelogUpdater() as updater:
    block = updater.changelog[0]
    if block.distributions == 'UNRELEASED':
        seen = set()
        for (author, linenos, change) in changes_by_author(block.changes()):
            change = ''.join(change)
            if (author, change) in seen:
                for lineno in reversed(linenos):
                    del block._changes[lineno]
            seen.add((author, change))


report_result('Remove duplicate line from changelog.')



#!/usr/bin/python3
import os
import sys

if not os.path.exists('debian/pyversions'):
    sys.exit(2)

with open('debian/pyversions', 'r') as f:
    pyversions = f.read().strip()

if pyversions.startswith('2.'):
    os.unlink('debian/pyversions')

print("Remove obsolete debian/pyversions.")
print("Fixed-Lintian-Tags: debian-pyversions-is-obsolete")



#!/usr/bin/python3

from lintian_brush.changelog import ChangelogUpdater, TextWrapper, WIDTH
from lintian_brush.fixer import report_result
import re


updated = []
initial_re = re.compile(r'^[  ]+[\+\-\*] ')


def rewrap_change(change):
    if not change:
        return change
    m = initial_re.match(change[0])
    if any([len(line) > WIDTH for line in change]) and m:
        wrapper = TextWrapper(m.group(0))
        return wrapper.wrap(
                '\n'.join(l[len(m.group(0)):] for l in change))
    else:
        return change


def rewrap_changes(changes):
    change = []
    indent = None
    for line in changes:
        m = initial_re.match(line)
        if m:
            yield from rewrap_change(change)
            change = [line]
            indent = len(m.group(0))
        elif change and line.startswith(' ' * indent):
            change.append(line)
        else:
            yield from rewrap_change(change)
            change = []
            yield line


with ChangelogUpdater() as updater:
    for block in updater.changelog:
        if not any([len(change) > WIDTH for change in block.changes()]):
            continue
        old_changes = list(block._changes)
        new_changes = rewrap_changes(block._changes)
        if old_changes != new_changes:
            block._changes = new_changes
            updated.append(block.version)

report_result(
    'Wrap long lines in changelog entries: %s.' % (
     ', '.join([str(v) for v in updated])),
    fixed_lintian_tags=['debian-changelog-line-too-long'])



#!/usr/bin/python3

from lintian_brush.control import (
    update_control,
    get_relation,
    drop_dependency,
    add_dependency,
    )


def move_libmodule_build_perl(control):
    try:
        get_relation(
            control.get('Build-Depends-Indep', ''), 'libmodule-build-perl')
    except KeyError:
        return

    control['Build-Depends-Indep'] = drop_dependency(
        control['Build-Depends-Indep'], 'libmodule-build-perl')
    control['Build-Depends'] = add_dependency(
        control.get('Build-Depends', ''), 'libmodule-build-perl')


update_control(source_package_cb=move_libmodule_build_perl)
print('Move libmodule-build-perl from Build-Depends-Indep to Build-Depends.')
print('Fixed: libmodule-build-perl-needs-to-be-in-build-depend')



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with control files."""

__all__ = ['guess_repository_url', 'determine_browser_url']

import re
from urllib.parse import urlparse

from .vcs import determine_gitlab_browser_url


MAINTAINER_EMAIL_MAP = {
    'pkg-javascript-devel@lists.alioth.debian.org': 'js-team',
    'python-modules-team@lists.alioth.debian.org': 'python-team/modules',
    'python-apps-team@lists.alioth.debian.org': 'python-team/applications',
    'debian-science-maintainers@lists.alioth.debian.org': 'science-team',
    'pkg-perl-maintainers@lists.alioth.debian.org':
        'perl-team/modules/packages',
    'pkg-java-maintainers@lists.alioth.debian.org': 'java-team',
    'pkg-ruby-extras-maintainers@lists.alioth.debian.org': 'ruby-team',
    'pkg-clamav-devel@lists.alioth.debian.org': 'clamav-team',
    'pkg-go-maintainers@lists.alioth.debian.org': 'go-team/packages',
    'pkg-games-devel@lists.alioth.debian.org': 'games-team',
    'pkg-telepathy-maintainers@lists.alioth.debian.org': 'telepathy-team',
    'debian-fonts@lists.debian.org': 'fonts-team',
    'pkg-gnustep-maintainers@lists.alioth.debian.org': 'gnustep-team',
    'pkg-gnome-maintainers@lists.alioth.debian.org': 'gnome-team',
    'pkg-multimedia-maintainers@lists.alioth.debian.org': 'multimedia-team',
    'debian-ocaml-maint@lists.debian.org': 'ocaml-team',
    'pkg-php-pear@lists.alioth.debian.org': 'php-team/pear',
    'pkg-mpd-maintainers@lists.alioth.debian.org': 'mpd-team',
    'pkg-cli-apps-team@lists.alioth.debian.org': 'dotnet-team',
    'pkg-mono-group@lists.alioth.debian.org': 'dotnet-team',
    }

TEAM_NAME_MAP = {
    'debian-xml-sgml': 'xml-sgml-team',
    'pkg-go': 'go-team',
    'pkg-fonts': 'fonts-team',
    'pkg-javascript': 'js-team',
    'pkg-java': 'java-team',
    'pkg-mpd': 'mpd-team',
    'pkg-electronics': 'electronics-team',
    'pkg-xfce': 'xfce-team',
    'pkg-lxc': 'lxc-team',
    'debian-science': 'science-team',
    'pkg-games': 'games-team',
    'pkg-bluetooth': 'bluetooth-team',
    'debichem': 'debichem-team',
    'openstack': 'openstack-team',
    'pkg-kde': 'qt-kde-team',
    'debian-islamic': 'islamic-team',
    'pkg-lua': 'lua-team',
    'pkg-xorg': 'xorg-team',
    'debian-astro': 'debian-astro-team',
    'pkg-icecast': 'multimedia-team',
    'glibc-bsd': 'bsd-team',
    'pkg-nvidia': 'nvidia-team',
    'pkg-llvm': 'llvm-team',
    'pkg-nagios': 'nagios-team',
    'pkg-sugar': 'pkg-sugar-team',
    'pkg-phototools': 'debian-phototools-team',
    'pkg-netmeasure': 'ineteng-team',
    'pkg-hamradio': 'debian-hamradio-team',
    'pkg-sass': 'sass-team',
    'pkg-rpm': 'pkg-rpm-team',
    'tts': 'tts-team',
    'python-apps': 'python-team/applications',
    'pkg-monitoring': 'monitoring-team',
    'pkg-perl': 'perl-team/modules',
    'debian-iot': 'debian-iot-team',
    'pkg-bitcoin': 'cryptocoin-team',
    'pkg-cyrus-imapd': 'debian',
    'pkg-dns': 'dns-team',
    'pkg-freeipa': 'freeipa-team',
    'pkg-ocaml-team': 'ocaml-team',
    'pkg-vdr-dvb': 'vdr-team',
    'debian-in': 'debian-in-team',
    'pkg-octave': 'pkg-octave-team',
    'pkg-postgresql': 'postgresql',
    'pkg-grass': 'debian-gis-team',
    'pkg-evolution': 'gnome-team',
    'pkg-gnome': 'gnome-team',
    'pkg-exppsy': 'neurodebian-team',
    'pkg-voip': 'pkg-voip-team',
    'pkg-privacy': 'pkg-privacy-team',
    'pkg-libvirt': 'libvirt-team',
    'debian-ha': 'ha-team',
    'debian-lego': 'debian-lego-team',
    'calendarserver': 'calendarserver-team',
    '3dprinter': '3dprinting-team',
    'pkg-multimedia': 'multimedia-team',
    'pkg-emacsen': 'emacsen-team',
    'pkg-haskell': 'haskell-team',
    'pkg-gnutls': 'gnutls-team',
    'pkg-mysql': 'mariadb-team',
    'pkg-php': 'php-team',
    'pkg-qemu': 'qemu-team',
    'pkg-xmpp': 'xmpp-team',
    'uefi': 'efi-team',
    'pkg-manpages-fr': 'l10n-fr-team',
    'pkg-proftpd': 'debian-proftpd-team',
    'pkg-apache': 'apache-team',
}

GIT_PATH_RENAMES = {
    'pkg-kde/applications': 'qt-kde-team/kde',
    '3dprinter/packages': '3dprinting-team',
    'pkg-emacsen/pkg': 'emacsen-team',
    'debian-astro/packages': 'debian-astro-team',
    'debian-islamic/packages': 'islamic-team',
    'debichem/packages': 'debichem-team',
    'pkg-privacy/packages': 'pkg-privacy-team',
}


def guess_repository_url(package, maintainer_email):
    """Guess the repository URL for a package hosted on Salsa.

    Args:
      package: Package name
      maintainer_email: The maintainer's email address (e.g. team list address)
    Returns:
      A guessed repository URL
    """
    if maintainer_email.endswith('@debian.org'):
        team_name = maintainer_email.split('@')[0]
    else:
        try:
            team_name = MAINTAINER_EMAIL_MAP[maintainer_email]
        except KeyError:
            return None

    return 'https://salsa.debian.org/%s/%s.git' % (team_name, package)


def determine_browser_url(url):
    """Determine the browser URL from a regular Git URL.

    Args:
      url: Regular URL
    Returns:
      a browser URL
    """
    return determine_gitlab_browser_url(url)


def _salsa_path_from_alioth_url(vcs_type, alioth_url):
    if vcs_type is None:
        return None
    # These two regular expressions come from vcswatch:
    # https://salsa.debian.org/qa/qa/blob/master/data/vcswatch/vcswatch#L165
    if vcs_type.lower() == 'git':
        m = ("(https?|git)://(anonscm|git).debian.org/"
             "(cgit/|git/)?collab-maint/")
        if re.match(m, alioth_url):
            return re.sub(m, 'debian/', alioth_url)
        m = "(https?|git)://(anonscm|git).debian.org/(cgit/|git/)?users/"
        if re.match(m, alioth_url):
            return re.sub(m, '', alioth_url)
        m = re.match(
            "(https?|git)://(anonscm|git).debian.org/(cgit/|git/)?(.+)",
            alioth_url)
        if m:
            parts = m.group(4).split('/')
            for i in range(len(parts), 0, -1):
                subpath = '/'.join(parts[:i])
                try:
                    return (
                        GIT_PATH_RENAMES[subpath] + '/' +
                        '/'.join(parts[i:]))
                except KeyError:
                    pass
        m = re.match(
            "(https?|git)://(anonscm|git).debian.org/(cgit/|git/)?([^/]+)/",
            alioth_url)
        if m and m.group(4) in TEAM_NAME_MAP:
            new_name = TEAM_NAME_MAP[m.group(4)]
            return re.sub(m.re, new_name + '/', alioth_url)
        m = re.match(
            'https?://alioth.debian.org/anonscm/(git/|cgit/)?([^/]+)/',
            alioth_url)
        if m and m.group(2) in TEAM_NAME_MAP:
            new_name = TEAM_NAME_MAP[m.group(2)]
            return re.sub(m.re, new_name + '/', alioth_url)

    if vcs_type.lower() == 'svn':
        if alioth_url.startswith('svn://svn.debian.org/pkg-perl/trunk'):
            return alioth_url.replace(
                'svn://svn.debian.org/pkg-perl/trunk',
                'perl-team/modules/packages')
        if alioth_url.startswith('svn://svn.debian.org/pkg-lua/packages'):
            return alioth_url.replace(
                'svn://svn.debian.org/pkg-lua/packages',
                'lua-team')
        parsed_url = urlparse(alioth_url)
        if (parsed_url.scheme == 'svn' and
                parsed_url.netloc in (
                    ('svn.debian.org', 'anonscm.debian.org'))):
            parts = parsed_url.path.strip('/').split('/')
            if parts[0] == 'svn':
                parts.pop(0)
            if (len(parts) == 3 and
                    parts[0] in TEAM_NAME_MAP and
                    parts[2] == 'trunk'):
                return '%s/%s' % (TEAM_NAME_MAP[parts[0]], parts[1])
            if (len(parts) == 3 and
                    parts[0] in TEAM_NAME_MAP and
                    parts[1] == 'trunk'):
                return '%s/%s' % (TEAM_NAME_MAP[parts[0]], parts[2])
            if (len(parts) == 4 and
                    parts[0] in TEAM_NAME_MAP and
                    parts[1] == 'packages' and
                    parts[3] == 'trunk'):
                return '%s/%s' % (TEAM_NAME_MAP[parts[0]], parts[2])
            if (len(parts) == 4 and
                    parts[0] in TEAM_NAME_MAP and
                    parts[1] == 'trunk' and parts[2] == 'packages'):
                return '%s/%s' % (TEAM_NAME_MAP[parts[0]], parts[3])
            if (len(parts) > 3 and
                    parts[0] in TEAM_NAME_MAP and
                    parts[-2] == 'trunk'):
                return '%s/%s' % (TEAM_NAME_MAP[parts[0]], parts[-1])
            if (len(parts) == 3 and
                    parts[0] in TEAM_NAME_MAP and
                    parts[1] in ('packages', 'unstable')):
                return '%s/%s' % (TEAM_NAME_MAP[parts[0]], parts[2])
    return None


def salsa_url_from_alioth_url(vcs_type, alioth_url):
    """Guess the salsa URL from an alioth URL.

    Args:
      vcs_type: VCS type
      alioth_url: Alioth URL
    Returns:
      Salsa URL
    """
    path = _salsa_path_from_alioth_url(vcs_type, alioth_url)
    if path is None:
        return None
    path = path.strip('/')
    if not path.endswith('.git'):
        path = path + '.git'
    return 'https://salsa.debian.org/' + path



#!/usr/bin/python3

from lintian_brush.control import update_control

FIXABLE_HOSTS = [
    'gitlab.com', 'github.com', 'salsa.debian.org',
    'gitorious.org']


def use_recommended_uri_format(control):
    if "Vcs-Git" in control:
        vcs_git = control["Vcs-Git"]
        if ':' not in vcs_git:
            return
        (netloc, path) = vcs_git.split(':', 1)
        if netloc.startswith('git@'):
            netloc = netloc[4:]
        if netloc in FIXABLE_HOSTS:
            control["Vcs-Git"] = 'https://%s/%s' % (netloc, path)


update_control(source_package_cb=use_recommended_uri_format)

print("Use recommended URI format in Vcs header.")
print("Fixed-Lintian-Tags: vcs-field-uses-not-recommended-uri-format")



#!/usr/bin/python3

import os

from lintian_brush import certainty_sufficient
from lintian_brush.control import ControlUpdater
from lintian_brush.upstream_metadata import guess_upstream_metadata_items

current_certainty = None
trust_package = os.environ.get('TRUST_PACKAGE') == 'true'

with ControlUpdater() as updater:
    if 'Homepage' not in updater.source:
        minimum_certainty = os.environ.get('MINIMUM_CERTAINTY')
        for datum in guess_upstream_metadata_items(
                '.', trust_package=trust_package):
            if datum.field != 'Homepage':
                continue
            if not certainty_sufficient(datum.certainty, minimum_certainty):
                continue
            if current_certainty != 'certain':
                updater.source["Homepage"] = datum.value
                current_certainty = datum.certainty

print('Fill in Homepage field.')
print('Fixed-Lintian-Tags: no-homepage-field')
if current_certainty:
    print('Certainty: %s' % current_certainty)



#!/usr/bin/python3

from lintian_brush.control import drop_dependency, ControlUpdater

with open('debian/rules', 'rb') as f:
    for line in f:
        if b'/usr/share/cdbs/' in line:
            uses_cdbs = True
            break
    else:
        uses_cdbs = False

if not uses_cdbs:
    with ControlUpdater() as updater:
        updater.source["Build-Depends"] = drop_dependency(
            updater.source.get("Build-Depends", ""), "cdbs")
        if not updater.source["Build-Depends"]:
            del updater.source["Build-Depends"]

print("Drop unused build-dependency on cdbs.")
print("Fixed-Lintian-Tags: unused-build-dependency-on-cdbs")



#!/usr/bin/python3

from debian.copyright import (
    LicenseParagraph,
    NotMachineReadableError,
    )
from lintian_brush.copyright import update_copyright

license = None


def fix_header_license_references(copyright):
    global license

    if not copyright.header.license:
        return
    if not copyright.header.license.text:
        return
    used_licenses = set()
    seen_licenses = set()
    for files_paragraph in copyright.all_files_paragraphs():
        if not files_paragraph.license:
            continue
        used_licenses.add(files_paragraph.license.synopsis)
        if files_paragraph.license.text:
            seen_licenses.add(files_paragraph.license.synopsis)
    for license_paragraph in copyright.all_license_paragraphs():
        seen_licenses.add(license_paragraph.license.synopsis)
    for missing in used_licenses - seen_licenses:
        if copyright.header.license.synopsis == missing:
            copyright.add_license_paragraph(
                LicenseParagraph.create(
                    copyright.header.license))
    license = copyright.header.license


try:
    update_copyright(fix_header_license_references)
except (FileNotFoundError, NotMachineReadableError):
    pass
else:
    if license:
        print('Add missing license paragraph for %s' % license.synopsis)
        print(
            'Fixed-Lintian-Tags: '
            'dep5-file-paragraph-references-header-paragraph')



#!/usr/bin/python3

from lintian_brush.control import (
    ControlUpdater,
    PkgRelation,
    parse_relations,
    add_dependency,
    is_relation_implied,
    )
from lintian_brush.lintian_overrides import override_exists
from lintian_brush.rules import Makefile, Rule, dh_invoke_get_with
import shlex
import sys

COMMAND_TO_DEP = {
    'dh_python2': 'dh-python | dh-sequence-python2',
    'dh_python3': 'dh-python | dh-sequence-python3',
    'dh_haskell_depends': 'haskell-devscripts-minimal',
    'dh_installtex': 'tex-common',
}

ADDON_TO_DEP = {
    'python3': 'dh-python | dh-sequence-python3',
    'autotools_dev': 'autotools-dev',
    'autoreconf':
        'dh-autoreconf | debhelper (>= 9.20160403~) | debhelper-compat',
    'bash_completion': 'bash-completion',
    'gir': 'gobject-introspection',
    'systemd': 'debhelper (>= 9.20160709~) | debhelper-compat | dh-systemd',
}

need = set()
tags = {}

mf = Makefile.from_path('debian/rules')

for entry in mf.contents:
    if isinstance(entry, Rule):
        for command in entry.commands():
            if command.lstrip().startswith(b'#'):
                continue
            try:
                executable = shlex.split(command.decode('utf-8', 'replace'))[0]
            except (ValueError, IndexError):
                continue
            try:
                dep = COMMAND_TO_DEP[executable]
            except KeyError:
                pass
            else:
                if override_exists(
                    'missing-build-dependency-for-dh_-command',
                        package='source', info='%s => %s' % (executable, dep)):
                    continue
                need.add(dep)
                tags[dep] = 'missing-build-dependency-for-dh_command'
            if executable == 'dh' or executable.startswith('dh_'):
                for addon in dh_invoke_get_with(command):
                    try:
                        dep = ADDON_TO_DEP[addon]
                    except KeyError:
                        pass
                    else:
                        need.add(dep)
                        tags[dep] = 'missing-build-dependency-for-dh-addon'


if not need:
    sys.exit(0)


fixed_tags = set()


with ControlUpdater() as updater:
    for deps in need:
        parsed = PkgRelation.parse(deps)
        build_deps = updater.source.get('Build-Depends', '')
        for unused1, existing, unused2 in parse_relations(build_deps):
            if is_relation_implied(parsed, existing):
                break
        else:
            updater.source['Build-Depends'] = add_dependency(build_deps, deps)
            fixed_tags.add(tags[deps])

print('Add missing build dependency on dh addon.')
print('Fixed-Lintian-Tags: %s' % ', '.join(fixed_tags))



#!/usr/bin/python3

# TODO(jelmer): Read python3 setup.py dist_info
# TODO(jelmer): Check XS-Go-Import-Path

from debian.changelog import Version
import os
import sys
from lintian_brush import (
    certainty_sufficient,
    min_certainty,
    )
from lintian_brush.fixer import net_access_allowed
from lintian_brush.upstream_metadata import (
    UpstreamDatum,
    check_upstream_metadata,
    extend_upstream_metadata,
    fix_upstream_metadata,
    guess_upstream_metadata_items,
    update_from_guesses,
    filter_bad_guesses,
    ADDON_ONLY_FIELDS,
    upstream_metadata_sort_key,
    upstream_version,
    )
from lintian_brush.yaml import (
    YamlUpdater,
    update_ordered_dict,
    )


current_version = Version(os.environ['CURRENT_VERSION'])


if not current_version.debian_revision:
    # Native package
    sys.exit(0)


with YamlUpdater('debian/upstream/metadata') as code:
    upstream_metadata = {
        k: UpstreamDatum(k, v, 'certain') for (k, v) in code.items()}

    minimum_certainty = os.environ.get('MINIMUM_CERTAINTY')
    trust_package = os.environ.get('TRUST_PACKAGE') == 'true'
    net_access = net_access_allowed()

    # Do some guessing based on what's in the package
    update_from_guesses(
        upstream_metadata, filter_bad_guesses(
            guess_upstream_metadata_items(
                '.', trust_package=trust_package)))

    # Then extend that by contacting e.g. SourceForge
    extend_upstream_metadata(
        upstream_metadata, '.',
        # Downgrade minimum certainty, since check_upstream_metadata can
        # upgrade it to "certain" later.
        minimum_certainty=(
            'likely'
            if net_access and minimum_certainty == 'certain'
            else minimum_certainty),
        net_access=net_access,
        consult_external_directory=True)
    if net_access:
        # Verify that online resources actually exist and adjust certainty
        # accordingly.
        upstream_version = upstream_version(current_version)
        check_upstream_metadata(upstream_metadata, version=upstream_version)

    # Homepage is set in debian/control, so don't add it to
    # debian/upstream/metadata.
    external_present_fields = set(['Homepage'])

    # If the debian/copyright file is machine-readable, then we do
    # not need to set the Name/Contact information in the
    # debian/upstream/metadata file.
    if 'Name' in upstream_metadata or 'Contact' in upstream_metadata:
        from lintian_brush.copyright import upstream_fields_in_copyright
        external_present_fields.update(upstream_fields_in_copyright().keys())

    for key, datum in list(upstream_metadata.items()):
        # Drop keys that don't need to be in debian/upsteam/metadata
        if key.startswith('X-') or key in external_present_fields:
            del upstream_metadata[key]

        # Drop everything that is below our minimum certainty
        elif not certainty_sufficient(datum.certainty, minimum_certainty):
            del upstream_metadata[key]

    achieved_certainty = min_certainty(
        [d.certainty for d in upstream_metadata.values()])

    fix_upstream_metadata(upstream_metadata)

    changed = {
        k: v
        for k, v in upstream_metadata.items()
        if v.value != code.get(k)}

    if not changed:
        sys.exit(0)

    # A change that just says the "Name" field is a bit silly
    if set(changed.keys()) - set(ADDON_ONLY_FIELDS) == set(['Name']):
        sys.exit(0)

    fixed_tag = not os.path.exists('debian/upstream/metadata')

    update_ordered_dict(
        code, [(k, v.value) for (k, v) in changed.items()],
        key=upstream_metadata_sort_key)

    # If there are only add-on-only fields, then just remove the file.
    if not (set(code.keys()) - set(ADDON_ONLY_FIELDS)):
        code.clear()

    if code and not os.path.isdir('debian/upstream'):
        os.makedirs('debian/upstream', exist_ok=True)


fields = [
    ('%s (from %s)' % (v.field, v.origin)) if v.origin else v.field
    for k, v in sorted(changed.items())
    ]

print('Set upstream metadata fields: %s.' % ', '.join(sorted(fields)))
print('Certainty: %s' % achieved_certainty)
if fixed_tag:
    print('Fixed-Lintian-Tags: upstream-metadata-file-is-missing')



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for applying multi-arch hints."""

import contextlib
import os
import re
import sys

from urllib.request import urlopen, Request

from lintian_brush import (
    Fixer,
    NoChanges,
    NotDebianPackage,
    PendingChanges,
    FixerResult,
    min_certainty,
    USER_AGENT,
    SUPPORTED_CERTAINTIES,
    DEFAULT_URLLIB_TIMEOUT,
    certainty_sufficient,
    get_committer,
    get_dirty_tracker,
    check_clean_tree,
    run_lintian_fixer,
    version_string,
    )
from .control import (
    update_control,
    format_relations,
    parse_relations,
    )


MULTIARCH_HINTS_URL = 'https://dedup.debian.net/static/multiarch-hints.yaml.xz'


def parse_multiarch_hints(f):
    """Parse a multi-arch hints file.

    Args:
      f: File-like object to read from
    Returns:
      dictionary mapping binary package names to lists of hints
    """
    import ruamel.yaml
    import ruamel.yaml.reader
    data = ruamel.yaml.load(f, ruamel.yaml.SafeLoader)
    if data.get('format') != 'multiarch-hints-1.0':
        raise ValueError('invalid file format: %r' % data.get('format'))
    return data['hints']


def multiarch_hints_by_binary(hints):
    ret = {}
    for entry in hints:
        ret.setdefault(entry['binary'], []).append(entry)
    return ret


def multiarch_hints_by_source(hints):
    ret = {}
    for entry in hints:
        if 'source' not in entry:
            continue
        ret.setdefault(entry['source'], []).append(entry)
    return ret


@contextlib.contextmanager
def download_multiarch_hints(url=MULTIARCH_HINTS_URL):
    """Load multi-arch hints from a URL.

    Args:
      url: URL to read from
    Returns:
      multi-arch hints file
    """
    headers = {'User-Agent': USER_AGENT}
    with urlopen(
            Request(url, headers=headers),
            timeout=DEFAULT_URLLIB_TIMEOUT) as f:
        if url.endswith('.xz'):
            import lzma
            # It would be nicer if there was a content-type, but there isn't
            # :-(
            f = lzma.LZMAFile(f)
        yield f


def apply_hint_ma_foreign(binary, hint):
    if binary.get('Multi-Arch') != 'foreign':
        binary['Multi-Arch'] = 'foreign'
        return 'Add Multi-Arch: foreign.'


def apply_hint_ma_foreign_lib(binary, hint):
    if binary.get('Multi-Arch') == 'foreign':
        del binary['Multi-Arch']
        return 'Drop Multi-Arch: foreign.'


def apply_hint_file_conflict(binary, hint):
    if binary.get('Multi-Arch') == 'same':
        del binary['Multi-Arch']
        return 'Drop Multi-Arch: same.'


def apply_hint_dep_any(binary, hint):
    m = re.match(
        '(.*) could have its dependency on (.*) annotated with :any',
        hint['description'])
    if not m or m.group(1) != binary['Package']:
        raise ValueError(
            'unable to parse hint description: %r' % hint['description'])
    dep = m.group(2)
    if 'Depends' not in binary:
        return
    changed = False
    relations = parse_relations(binary['Depends'])
    for entry in relations:
        (head_whitespace, relation, tail_whitespace) = entry
        if not isinstance(relation, str):  # formatting
            for r in relation:
                if r.name == dep and r.archqual != 'any':
                    r.archqual = 'any'
                    changed = True
    if not changed:
        return
    binary['Depends'] = format_relations(relations)
    return ('Add :all qualifier for %s dependency.' % dep)


def apply_hint_ma_same(binary, hint):
    if binary.get('Multi-Arch') == 'same':
        return
    binary['Multi-Arch'] = 'same'
    return 'Add Multi-Arch: same.'


def apply_hint_arch_all(binary, hint):
    if binary['Architecture'] == 'all':
        return
    binary['Architecture'] = 'all'
    return 'Make package Architecture: all.'


class MultiArchHintApplier(object):

    def __init__(self, kind, fn, certainty):
        self.kind = kind
        self.fn = fn
        self.certainty = certainty


class MultiArchFixerResult(FixerResult):

    def __init__(self, description, certainty, changes):
        super(MultiArchFixerResult, self).__init__(
            description=description, certainty=certainty)
        self.changes = changes


def apply_multiarch_hints(hints, minimum_certainty='certain'):
    changes = []
    appliers = {applier.kind: applier for applier in APPLIERS}

    def update_binary(binary):
        for hint in hints.get(binary['Package'], []):
            kind = hint['link'].rsplit('#', 1)[1]
            applier = appliers[kind]
            if not certainty_sufficient(
                    applier.certainty, minimum_certainty):
                continue
            description = applier.fn(binary, hint)
            if description:
                changes.append(
                    (binary, hint, description, applier.certainty))
    update_control(binary_package_cb=update_binary)
    return changes


class MultiArchHintFixer(Fixer):

    def __init__(self, hints):
        super(MultiArchHintFixer, self).__init__(name='multiarch-hints')
        self._hints = hints

    def run(self, basedir, package, current_version, compat_release,
            minimum_certainty=None, trust_package=False,
            allow_reformatting=False, net_access=True, opinionated=False,
            diligence=0):
        if not net_access:
            # This should never happen - perhaps if something else imported and
            # used this class?
            raise NoChanges()
        old_cwd = os.getcwd()
        try:
            os.chdir(basedir)
            changes = apply_multiarch_hints(self._hints, minimum_certainty)
        finally:
            os.chdir(old_cwd)

        overall_certainty = min_certainty(
            [certainty for (binary, hint, description, certainty) in changes])
        overall_description = "Apply multi-arch hints.\n\n" + "\n".join(
            ["* %s: %s" % (binary['Package'], description)
             for (binary, hint, description, certainty) in changes])
        return MultiArchFixerResult(
            overall_description, certainty=overall_certainty, changes=changes)


APPLIERS = [
    MultiArchHintApplier('ma-foreign', apply_hint_ma_foreign, 'certain'),
    MultiArchHintApplier('file-conflict', apply_hint_file_conflict, 'certain'),
    MultiArchHintApplier(
        'ma-foreign-library', apply_hint_ma_foreign_lib, 'certain'),
    MultiArchHintApplier('dep-any', apply_hint_dep_any, 'certain'),
    MultiArchHintApplier('ma-same', apply_hint_ma_same, 'certain'),
    MultiArchHintApplier('arch-all', apply_hint_arch_all, 'possible'),
]


def main(argv=None):
    import argparse
    import sys
    from breezy.workingtree import WorkingTree
    import locale
    locale.setlocale(locale.LC_ALL, '')
    # Use better default than ascii with posix filesystems that deal in bytes
    # natively even when the C locale or no locale at all is given. Note that
    # we need an immortal string for the hack, hence the lack of a hyphen.
    sys._brz_default_fs_enc = "utf8"

    import breezy  # noqa: E402
    breezy.initialize()
    import breezy.git  # noqa: E402
    import breezy.bzr  # noqa: E402
    from breezy.trace import note  # noqa: E402

    from .config import Config

    parser = argparse.ArgumentParser(prog='multi-arch-fixer')
    parser.add_argument(
        '--directory', metavar='DIRECTORY', help='directory to run in',
        type=str, default='.')
    parser.add_argument(
        '--disable-inotify', action='store_true', default=False,
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--identity',
        help='Print user identity that would be used when committing',
        action='store_true', default=False)
    # Hide the minimum-certainty option for the moment.
    parser.add_argument(
        '--minimum-certainty',
        type=str,
        choices=SUPPORTED_CERTAINTIES,
        default=None,
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--no-update-changelog', action="store_false", default=None,
        dest="update_changelog", help="do not update the changelog")
    parser.add_argument(
        '--update-changelog', action="store_true", dest="update_changelog",
        help="force updating of the changelog", default=None)
    parser.add_argument(
        '--version', action='version', version='%(prog)s ' + version_string)
    parser.add_argument(
        '--allow-reformatting', default=None, action='store_true',
        help=argparse.SUPPRESS)

    args = parser.parse_args(argv)
    minimum_certainty = args.minimum_certainty
    wt, subpath = WorkingTree.open_containing(args.directory)
    if args.identity:
        note(get_committer(wt))
        return 0

    update_changelog = args.update_changelog
    allow_reformatting = args.allow_reformatting
    try:
        cfg = Config.from_workingtree(wt, subpath)
    except FileNotFoundError:
        pass
    else:
        if minimum_certainty is None:
            minimum_certainty = cfg.minimum_certainty()
        if allow_reformatting is None:
            allow_reformatting = cfg.allow_reformatting()
        if update_changelog is None:
            update_changelog = cfg.update_changelog()

    use_inotify = (False if args.disable_inotify else None),
    try:
        check_clean_tree(wt)
    except PendingChanges:
        note("%s: Please commit pending changes first.", wt.basedir)
        return 1

    dirty_tracker = get_dirty_tracker(wt, subpath, use_inotify)
    if dirty_tracker:
        dirty_tracker.mark_clean()

    note("Downloading multiarch hints.")
    with download_multiarch_hints() as f:
        hints = multiarch_hints_by_binary(parse_multiarch_hints(f))

    try:
        result, summary = run_lintian_fixer(
            wt, MultiArchHintFixer(hints),
            update_changelog=update_changelog,
            minimum_certainty=minimum_certainty,
            dirty_tracker=dirty_tracker,
            subpath=subpath, allow_reformatting=allow_reformatting,
            net_access=True)
    except NoChanges:
        note('Nothing to do.')
    except NotDebianPackage:
        note("%s: Not a debian package.", wt.basedir)
        return 1
    else:
        for binary, hint, description, certainty in result.changes:
            note('%s: %s' % (binary['Package'], description))


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.patches."""

from io import BytesIO

from breezy.tests import (
    TestCase,
    TestCaseWithTransport,
    )
from breezy.patches import (
    parse_patch,
    Patch,
    )

from ..patches import (
    AppliedPatches,
    find_common_patch_suffix,
    find_patch_base,
    find_patches_branch,
    read_quilt_patches,
    read_quilt_series,
    tree_non_patches_changes,
    tree_patches_directory,
    upstream_with_applied_patches,
    )


class ReadSeriesFileTests(TestCase):

    def test_comment(self):
        f = BytesIO(b"""\
# This file intentionally left blank.
""")
        self.assertEqual(
            [('This', True, ['file', 'intentionally', 'left', 'blank.'])],
            list(read_quilt_series(f)))

    def test_empty(self):
        f = BytesIO(b"\n")
        self.assertEqual([], list(read_quilt_series(f)))

    def test_empty_line(self):
        f = BytesIO(b"""\
patch1

patch2
""")
        self.assertEqual(
            [
                ("patch1", False, []),
                ("patch2", False, []),
            ], list(read_quilt_series(f)))

    def test_options(self):
        f = BytesIO(b"""\
name -p0
name2
""")
        self.assertEqual(
            [
                ("name", False, ["-p0"]),
                ("name2", False, []),
            ],
            list(read_quilt_series(f)))


class FindPatchBaseTests(TestCaseWithTransport):

    def setUp(self):
        super(FindPatchBaseTests, self).setUp()
        self.tree = self.make_branch_and_tree('.')
        self.upstream_revid = self.tree.commit('upstream')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/changelog', """\
blah (0.38) unstable; urgency=medium

  * Fix something

 -- Jelmer Vernooij <jelmer@debian.org>  Sat, 19 Oct 2019 15:21:53 +0000
 """)])
        self.tree.add(['debian', 'debian/changelog'])

    def test_none(self):
        self.assertEqual(None, find_patch_base(self.tree))

    def test_upstream_dash(self):
        self.tree.branch.tags.set_tag('upstream-0.38', self.upstream_revid)
        self.assertEqual(
            self.upstream_revid, find_patch_base(self.tree))


class FindPatchBranchTests(TestCaseWithTransport):

    def make_named_branch_and_tree(self, path, name):
        dir = self.make_controldir(path)
        dir.create_repository()
        branch = dir.create_branch(name=name)
        dir.set_branch_reference(branch)
        return dir.create_workingtree()

    def test_none(self):
        tree = self.make_branch_and_tree('.')
        self.assertIs(None, find_patches_branch(tree))

    def test_patch_queue(self):
        master = self.make_named_branch_and_tree('.', name='master')
        master.branch.controldir.create_branch(name='patch-queue/master')

        self.assertEqual(
            'patch-queue/master',
            find_patches_branch(master).name)

    def test_patched_master(self):
        master = self.make_named_branch_and_tree('.', name='master')
        master.branch.controldir.create_branch(name='patched')
        self.assertEqual('patched', find_patches_branch(master).name)

    def test_patched_other(self):
        other = self.make_named_branch_and_tree('.', name='other')
        other.branch.controldir.create_branch(name='patched-other')
        self.assertEqual('patched-other', find_patches_branch(other).name)


class AppliedPatchesTests(TestCaseWithTransport):

    def test_apply_simple(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents([('a', 'a\n')])
        tree.add('a')
        tree.commit('Add a')
        patch = parse_patch(b"""\
--- a/a
+++ b/a
@@ -1 +1 @@
-a
+b
""".splitlines(True))
        with AppliedPatches(tree, [patch]) as newtree:
            self.assertEqual(b'b\n', newtree.get_file_text('a'))

    def test_apply_delete(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents([('a', 'a\n')])
        tree.add('a')
        tree.commit('Add a')
        patch = parse_patch(b"""\
--- a/a
+++ /dev/null
@@ -1 +0,0 @@
-a
""".splitlines(True))
        with AppliedPatches(tree, [patch]) as newtree:
            self.assertFalse(newtree.has_filename('a'))

    def test_apply_add(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents([('a', 'a\n')])
        tree.add('a')
        tree.commit('Add a')
        patch = parse_patch(b"""\
--- /dev/null
+++ b/b
@@ -0,0 +1 @@
+b
""".splitlines(True))
        with AppliedPatches(tree, [patch]) as newtree:
            self.assertEqual(b'b\n', newtree.get_file_text('b'))


class ReadQuiltPatchesTests(TestCaseWithTransport):

    def test_read_patches(self):
        patch = """\
--- a/a
+++ b/a
@@ -1,5 +1,5 @@
 line 1
 line 2
-line 3
+new line 3
 line 4
 line 5
"""
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/patches/', ),
            ('debian/patches/series', 'foo\n'),
            ('debian/patches/foo', patch)])
        tree.add(['debian', 'debian/patches', 'debian/patches/series',
                  'debian/patches/foo'])
        tree.commit('add patch')
        patches = list(read_quilt_patches(tree))
        self.assertEqual(1, len(patches))
        self.assertIsInstance(patches[0], Patch)
        self.assertEqual(patch.encode('utf-8'), patches[0].as_bytes())

    def test_no_series_file(self):
        t = self.make_branch_and_tree('.')
        self.assertEqual([], list(read_quilt_patches(t)))

    def test_comments(self):
        t = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/patches/', ),
            ('debian/patches/series',
             "# This file intentionally left blank.\n")])
        self.assertEqual([], list(read_quilt_patches(t)))


class UpstreamWithAppliedPatchesTests(TestCaseWithTransport):

    def setUp(self):
        super(UpstreamWithAppliedPatchesTests, self).setUp()
        self.tree = self.make_branch_and_tree('.')
        self.build_tree_contents([('afile', 'some line\n')])
        self.tree.add('afile')
        self.upstream_revid = self.tree.commit('upstream')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/changelog', """\
blah (0.38) unstable; urgency=medium

  * Fix something

 -- Jelmer Vernooij <jelmer@debian.org>  Sat, 19 Oct 2019 15:21:53 +0000
 """),
            ('debian/patches/', ),
            ('debian/patches/series', '1.patch\n'),
            ('debian/patches/1.patch', """\
--- a/afile
+++ b/afile
@@ -1 +1 @@
-some line
+another line
--- /dev/null
+++ b/newfile
@@ -0,0 +1 @@
+new line
"""),
            ('unchangedfile', 'unchanged\n')])
        self.tree.add(['debian', 'debian/changelog', 'unchangedfile'])

    def test_upstream_branch(self):
        self.tree.branch.tags.set_tag('upstream/0.38', self.upstream_revid)
        patches = list(read_quilt_patches(self.tree))
        with upstream_with_applied_patches(self.tree, patches) as t:
            self.assertEqual(b'another line\n', t.get_file_text('afile'))
            self.assertEqual(b'new line\n', t.get_file_text('newfile'))
            # TODO(jelmer): PreviewTree appears to be broken
            # self.assertEqual(b'unchanged\n',
            #                  t.get_file_text('unchangedfile'))


class TreePatchesNonPatchesTests(TestCaseWithTransport):

    def setUp(self):
        super(TreePatchesNonPatchesTests, self).setUp()
        self.tree = self.make_branch_and_tree('.')
        self.build_tree_contents([('afile', 'some line\n')])
        self.tree.add('afile')
        self.upstream_revid = self.tree.commit('upstream')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/changelog', """\
blah (0.38) unstable; urgency=medium

  * Fix something

 -- Jelmer Vernooij <jelmer@debian.org>  Sat, 19 Oct 2019 15:21:53 +0000
 """),
            ('debian/patches/', ),
            ('debian/patches/series', '1.patch\n'),
            ('debian/patches/1.patch', """\
--- a/afile
+++ b/afile
@@ -1 +1 @@
-some line
+another line
""")])
        self.tree.add(['debian', 'debian/changelog'])

    def test_no_delta(self):
        self.tree.branch.tags.set_tag('upstream/0.38', self.upstream_revid)
        self.assertEqual([], list(tree_non_patches_changes(
            self.tree, 'debian/patches')))

    def test_delta(self):
        self.tree.branch.tags.set_tag('upstream/0.38', self.upstream_revid)
        self.build_tree_contents([('anotherfile', 'blah')])
        self.tree.add('anotherfile')
        self.assertEqual(1, len(list(tree_non_patches_changes(
            self.tree, 'debian/patches'))))


class FindCommonPatchSuffixTests(TestCase):

    def test_simple(self):
        self.assertEqual(
            '.blah',
            find_common_patch_suffix(['foo.blah', 'series']))
        self.assertEqual(
            '.patch', find_common_patch_suffix(['foo.patch', 'series']))
        self.assertEqual(
            '.patch', find_common_patch_suffix(['series']))
        self.assertEqual(
            '.blah', find_common_patch_suffix(['series'], '.blah'))
        self.assertEqual(
            '', find_common_patch_suffix(['series', 'foo', 'bar']))
        self.assertEqual(
            '.patch',
            find_common_patch_suffix(['series', 'foo.patch', 'bar.patch']))


class TreePatchesDirectoryTests(TestCaseWithTransport):

    def test_none(self):
        tree = self.make_branch_and_tree('.')
        self.assertIs(None, tree_patches_directory(tree))

    def test_default(self):
        tree = self.make_branch_and_tree('.')
        tree.mkdir('debian')
        tree.mkdir('debian/patches')
        self.assertEqual('debian/patches', tree_patches_directory(tree))

    def test_custom(self):
        tree = self.make_branch_and_tree('.')
        tree.mkdir('debian')
        tree.mkdir('debian/patches-applied')
        self.build_tree_contents([('debian/rules', """\

export QUILT_PATCH_DIR := debian/patches-applied

all:

blah: bloe
\tfoo
""")])
        tree.add('debian/rules')
        self.assertEqual(
            'debian/patches-applied', tree_patches_directory(tree))



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Track whether a particular directory structure is dirty."""

import os
from pyinotify import (
    WatchManager,
    IN_CREATE,
    IN_CLOSE_WRITE,
    IN_Q_OVERFLOW,
    IN_DELETE,
    IN_MOVED_TO,
    IN_MOVED_FROM,
    IN_ATTRIB,
    ProcessEvent,
    Notifier,
    )


MASK = (
    IN_CLOSE_WRITE | IN_DELETE | IN_Q_OVERFLOW | IN_MOVED_TO | IN_MOVED_FROM |
    IN_ATTRIB)


class _Process(ProcessEvent):

    def my_init(self):
        self.paths = set()
        self.created = set()

    def process_default(self, event):
        path = os.path.join(event.path, event.name)
        if event.mask & IN_CREATE:
            self.created.add(path)
        self.paths.add(path)
        if event.mask & IN_DELETE and path in self.created:
            self.paths.remove(path)
            self.created.remove(path)


class DirtyTracker(object):

    def __init__(self, tree, subpath='.'):
        self._tree = tree
        self._wm = WatchManager()
        self._process = _Process()
        self._notifier = Notifier(self._wm, self._process)
        self._notifier.coalesce_events(True)

        def check_excluded(p):
            return tree.is_control_filename(tree.relpath(p))
        self._wdd = self._wm.add_watch(
            tree.abspath(subpath), MASK, rec=True, auto_add=True,
            exclude_filter=check_excluded)

    def _process_pending(self):
        if self._notifier.check_events(timeout=0):
            self._notifier.read_events()
        self._notifier.process_events()

    def mark_clean(self):
        self._process_pending()
        self._process.paths.clear()
        self._process.created.clear()

    def is_dirty(self):
        self._process_pending()
        return bool(self._process.paths)

    def paths(self):
        self._process_pending()
        return self._process.paths

    def relpaths(self):
        return set(self._tree.relpath(p) for p in self.paths())



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Lintian-brush configuration file."""

import os

from configobj import ConfigObj
import warnings
import distro_info


PACKAGE_CONFIG_FILENAME = 'debian/lintian-brush.conf'


SUPPORTED_KEYS = [
    'compat-release',
    'minimum-certainty',
    'allow-reformatting',
    'update-changelog',
    ]


def resolve_release_codename(name):
    debian = distro_info.DebianDistroInfo()
    if debian.codename(name):
        return debian.codename(name)
    if debian.valid(name):
        return name
    ubuntu = distro_info.UbuntuDistroInfo()
    if ubuntu.valid(name):
        return name
    return None


class Config(object):
    """A configuration file."""

    def __init__(self, path):
        if not os.path.exists(path):
            raise FileNotFoundError(path)
        self._obj = ConfigObj(path, raise_errors=True, file_error=True)
        for k in self._obj.keys():
            if k not in SUPPORTED_KEYS:
                warnings.warn('unknown setting %s in %s' % (k, path))

    @classmethod
    def from_workingtree(cls, wt, subpath):
        return cls(os.path.join(wt.basedir, subpath, PACKAGE_CONFIG_FILENAME))

    def compat_release(self):
        value = self._obj.get('compat-release')
        codename = resolve_release_codename(value)
        if codename is None:
            warnings.warn('unknown compat release %s, ignoring.' % value)
        return codename

    def allow_reformatting(self):
        try:
            return self._obj.as_bool('allow-reformatting')
        except KeyError:
            return None

    def minimum_certainty(self):
        return self._obj.get('minimum-certainty')

    def update_changelog(self):
        try:
            return self._obj.as_bool('update-changelog')
        except KeyError:
            return None



#!/usr/bin/python3

from lintian_brush.reformatting import check_generated_file

check_generated_file('debian/rules')

with open('debian/rules', 'rb') as f:
    oldcontents = f.read()

newcontents = oldcontents.replace(b'$(PWD)', b'$(CURDIR)')

if oldcontents != newcontents:
    with open('debian/rules', 'wb') as f:
        f.write(newcontents)

print("debian/rules: Avoid using $(PWD) variable.")
print("Fixed-Lintian-Tags: debian-rules-should-not-use-pwd")



#!/usr/bin/python3

import re
from lintian_brush.rules import update_rules, Makefile

architecture_included = False
PATH = '/usr/share/dpkg/architecture.mk'
INCLUDE_LINE = b'include ' + PATH.encode('ascii')

_variables = set()


def variable_defined(var):
    global _variables
    if _variables:
        return var in _variables
    k = {}
    with open(PATH, 'r') as f:
        for l in f:
            if l.strip().startswith('$(foreach '):
                vs = l.strip()[len('$(foreach '):].split(',')
                k[vs[0]] = vs[1]
    for machine in k['machine'].split(' '):
        for var in k['var'].split(' '):
            _variables.add('DEB_%s_%s' % (machine, var))


def is_dpkg_architecture_line(line):
    m = re.match(b'([A-Z_]+)[ \t]*[:?]?=[ \t]*(.*)', line.strip())
    if not m:
        return (False, False)
    variable = m.group(1)
    assignment = m.group(2)
    if not variable_defined(variable.decode()):
        return (False, False)
    m = re.match(b'\\$\\(shell dpkg-architecture -q([A-Z_]+)\\)', assignment)
    if not m or variable != m.group(1):
        return (True, False)
    else:
        return (True, True)


def drop_arch_line(line):
    global architecture_included
    if line.strip() == INCLUDE_LINE:
        architecture_included = True
        return line
    (variable_matches, value_matches) = is_dpkg_architecture_line(line)
    if not variable_matches:
        return line
    if not value_matches:
        # The assignment is different; let's include
        # the mk file *before* this assignment
        if not architecture_included and variable_matches:
            architecture_included = True
            return [INCLUDE_LINE, line]
        else:
            return line
    if not architecture_included:
        architecture_included = True
        return INCLUDE_LINE
    return None


try:
    mf = Makefile.from_path('debian/rules')
except FileNotFoundError:
    pass
else:
    for line in mf.contents:
        if (isinstance(line, bytes) and
                is_dpkg_architecture_line(line) == (True, True)):
            update_rules(global_line_cb=drop_arch_line)
            break
print('Rely on pre-initialized dpkg-architecture variables.')
print('Fixed-Lintian-Tags: debian-rules-sets-dpkg-architecture-variable')



#!/usr/bin/python3


from lintian_brush.copyright import update_copyright, NotMachineReadableError


def swap_files_glob(copyright):
    files_i = 0
    for i, paragraph in enumerate(copyright._Copyright__paragraphs):
        if "Files" in paragraph:
            if paragraph["Files"] == "*" and files_i > 0:
                copyright._Copyright__paragraphs.insert(
                    0, copyright._Copyright__paragraphs.pop(i))
            files_i += 1


try:
    update_copyright(swap_files_glob)
except (FileNotFoundError, NotMachineReadableError):
    pass

print('Make "Files: *" paragraph the first in the copyright file.')
print('Fixed-Lintian-Tags: '
      'global-files-wildcard-not-first-paragraph-in-dep5-copyright')



#!/usr/bin/python3

from lintian_brush.rules import update_rules


def update_line(line, target):
    if line.strip() == b'dh_clean -k':
        return b'dh_prep'
    return line


update_rules(update_line)
print("debian/rules: Use dh_prep rather than \"dh_clean -k\".")
print("Fixed-Lintian-Tags: dh-clean-k-is-deprecated")



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Debhelper utility functions."""


import os
import subprocess


DEBHELPER_BUILD_STEPS = ['configure', 'build', 'test', 'install', 'clean']


def detect_debhelper_buildsystem(step=None):
    if os.path.exists('configure.ac') or os.path.exists('configure.in'):
        return 'autoconf'
    output = subprocess.check_output([
        'perl', '-w', '-MDebian::Debhelper::Dh_Lib',
        '-MDebian::Debhelper::Dh_Buildsystems', '-e',
        """\
Debian::Debhelper::Dh_Lib::init();
my $b=Debian::Debhelper::Dh_Buildsystems::load_buildsystem(undef, %(step)s);\
if (defined($b)) { print($b->NAME); } else { print("_undefined_"); }\
""" % {"step": ("'%s'" % step) if step is not None else 'undef'}]).decode()
    if output == '_undefined_':
        return None
    return output


def lowest_non_deprecated_compat_level():
    output = subprocess.check_output([
        'perl', '-w', '-MDebian::Debhelper::Dh_Lib', '-e',
        'print(Debian::Debhelper::Dh_Lib::LOWEST_NON_DEPRECATED_COMPAT_LEVEL);'
        ]).decode()
    return int(output)


debhelper_compat_version = {
    # Debian
    'jessie': 9,
    'stretch': 10,
    'sid': 12,
    'bullseye': 12,
    'buster': 12,

    # Ubuntu
    'xenial': 9,
    'bionic': 11,
    'cosmic': 11,
    'disco': 12,
    'eoan': 12,
    }


def maximum_debhelper_compat_version(compat_release):
    max_version = debhelper_compat_version.get(compat_release)
    if max_version is None:
        max_version = lowest_non_deprecated_compat_level()
    return max_version



#!/usr/bin/python3

from lintian_brush.rules import update_rules

certainty = None


def drop_get_orig_source(rule):
    global certainty
    if rule.has_target(b'get-orig-source'):
        commands = rule.commands()
        if [b'uscan'] == [c.split(b' ')[0] for c in commands]:
            certainty = 'certain'
        else:
            certainty = 'possible'
        rule.clear()


update_rules(rule_cb=drop_get_orig_source)
print('Remove unnecessary get-orig-source-target.')
print('Fixed-Lintian-Tags: '
      'debian-rules-contains-unnecessary-get-orig-source-target')
if certainty:
    print('Certainty: %s' % certainty)
